Package: csdb
Title: An Abstracted System for Easily Working with Databases with
        Large Datasets
Version: 2026.5.13
Authors@R: c(
  person(c("Richard","Aubrey"), "White", email = "hello@rwhite.no", role = c("aut", "cre"), comment = c(ORCID = "0000-0002-6747-1726")),
  person(c("August", "Sørli"), "Mathisen", email = "august.sorli.mathisen@fhi.no", role = c("aut"))
  )
Description: Provides object-oriented database management tools for working with large datasets across multiple database systems. Features include robust connection management for 'PostgreSQL' databases, advanced table operations with bulk data loading and upsert functionality, comprehensive data validation through customizable field type and content validators, efficient index management, and cross-database compatibility. Designed for high-performance data operations in surveillance systems and large-scale data processing workflows.
Depends: R (>= 4.1.0)
License: MIT + file LICENSE
URL: https://niphr.github.io/csdb/, https://github.com/niphr/csdb
BugReports: https://github.com/niphr/csdb/issues
Encoding: UTF-8
LazyData: true
Imports: csutil, data.table, DBI, dplyr, fs, ggplot2, glue, methods,
        odbc, R6, S7, stringr, uuid
Suggests: testthat, knitr, rmarkdown, rstudioapi, digest, crayon
RoxygenNote: 7.3.3
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-05-13 04:59:48 UTC; raw996
Author: Richard Aubrey White [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-6747-1726>),
  August Sørli Mathisen [aut]
Maintainer: Richard Aubrey White <hello@rwhite.no>
Repository: CRAN
Date/Publication: 2026-05-13 05:20:09 UTC
Built: R 4.5.2; ; 2026-05-13 09:58:33 UTC; unix

Package: csdb
Title: An Abstracted System for Easily Working with Databases with
        Large Datasets
Version: 2026.5.13
Authors@R: c(
  person(c("Richard","Aubrey"), "White", email = "hello@rwhite.no", role = c("aut", "cre"), comment = c(ORCID = "0000-0002-6747-1726")),
  person(c("August", "Sørli"), "Mathisen", email = "august.sorli.mathisen@fhi.no", role = c("aut"))
  )
Description: Provides object-oriented database management tools for working with large datasets across multiple database systems. Features include robust connection management for 'PostgreSQL' databases, advanced table operations with bulk data loading and upsert functionality, comprehensive data validation through customizable field type and content validators, efficient index management, and cross-database compatibility. Designed for high-performance data operations in surveillance systems and large-scale data processing workflows.
Depends: R (>= 4.1.0)
License: MIT + file LICENSE
URL: https://niphr.github.io/csdb/, https://github.com/niphr/csdb
BugReports: https://github.com/niphr/csdb/issues
Encoding: UTF-8
LazyData: true
Imports: csutil, data.table, DBI, dplyr, fs, ggplot2, glue, methods,
        odbc, R6, S7, stringr, uuid
Suggests: testthat, knitr, rmarkdown, rstudioapi, digest, crayon
RoxygenNote: 7.3.3
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-05-13 04:59:48 UTC; raw996
Author: Richard Aubrey White [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-6747-1726>),
  August Sørli Mathisen [aut]
Maintainer: Richard Aubrey White <hello@rwhite.no>
Repository: CRAN
Date/Publication: 2026-05-13 05:20:09 UTC
Built: R 4.6.0; ; 2026-05-13 05:23:55 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(get_table_names_and_info,"Microsoft SQL Server")
S3method(get_table_names_and_info,PostgreSQL)
export(DBConnection_v9)
export(DBTable_v9)
export(csdb_get_auth_hook)
export(csdb_set_auth_hook)
export(get_table_names_and_info)
export(validator_field_contents_blank)
export(validator_field_contents_csfmt_rts_data_v1)
export(validator_field_contents_csfmt_rts_data_v2)
export(validator_field_types_blank)
export(validator_field_types_csfmt_rts_data_v1)
export(validator_field_types_csfmt_rts_data_v2)
import(R6)
import(data.table)
import(ggplot2)
importFrom(methods,getClass)
importFrom(methods,initialize)
importFrom(methods,isClass)
importFrom(stats,runif)

# csdb <a href="https://niphr.github.io/csdb/"><img src="man/figures/logo.png" align="right" width="120" /></a>

[![CRAN status](https://www.r-pkg.org/badges/version/csdb)](https://cran.r-project.org/package=csdb)
[![CRAN downloads](https://cranlogs.r-pkg.org/badges/csdb)](https://cran.r-project.org/package=csdb)

## Overview 

[csdb](https://niphr.github.io/csdb/) provides an abstracted system for easily working with databases with large datasets.

Read the introduction vignette [here](https://niphr.github.io/csdb/articles/csdb.html) or run `help(package="csdb")`.

# General utility functions for csdb package

#' Generate a random UUID string
#'
#' Creates a random UUID string suitable for use as temporary identifiers.
#' The UUID is modified to ensure it starts with a letter and includes
#' additional randomness for seed compatibility.
#'
#' @return A character string containing a random UUID
#' @keywords internal
#' @noRd
random_uuid <- function() {
  x <- uuid::UUIDgenerate(F)
  x <- gsub("-", "", x)
  # the second part here allows for the usage of set.seed()
  x <- paste0("a", x, round(runif(1)*10000000))
  x
}

#' Generate a random file path
#'
#' Creates a random file path in the specified folder with the given extension.
#' The folder is created if it doesn't exist.
#'
#' @param folder Directory path where the file should be created
#' @param extension File extension to use (default: ".csv")
#' @param extra_insert Additional string to insert in filename (optional)
#' @return A character string containing the full file path
#' @keywords internal
#' @noRd
random_file <- function(folder, extension = ".csv", extra_insert = NULL) {
  dir.create(folder, showWarnings = FALSE, recursive = TRUE)
  fs::path(folder, paste0(random_uuid(), extra_insert, extension))
}

# odbc isn't getting recognized within the R6 classes
# and the CRAN checks are therefore throwing an error
fix_r6 <- function(){
  odbc::odbc()
}

# dplyr and csutil isn't getting recognized within the S7 methods
# and the CRAN checks are therefore throwing an error
fix_s7 <- function(){
  csutil::easy_split()
  dplyr::filter()
}

#' Blank field types validator
#' 
#' A pass-through validator that accepts any field types without validation.
#' This is useful as a placeholder when no specific field type validation is needed.
#' 
#' @param db_field_types A named character vector of database field types
#' @return Always returns TRUE
#' @export
#' @examples
#' # This validator always returns TRUE regardless of input
#' field_types <- c("id" = "INTEGER", "name" = "TEXT", "date" = "DATE")
#' validator_field_types_blank(field_types)
#' 
#' # Works with any field types
#' other_types <- c("value" = "DOUBLE", "status" = "BOOLEAN")
#' validator_field_types_blank(other_types)
validator_field_types_blank <- function(db_field_types) {
  return(TRUE)
}

#' Blank data contents validator
#' 
#' A pass-through validator that accepts any data without validation.
#' This is useful as a placeholder when no specific data content validation is needed.
#' 
#' @param data A data.frame or data.table containing the data to validate
#' @return Always returns TRUE
#' @export
#' @examples
#' # This validator always returns TRUE regardless of input
#' test_data <- data.frame(id = 1:3, name = c("A", "B", "C"), value = c(10, 20, 30))
#' validator_field_contents_blank(test_data)
#' 
#' # Works with any data structure
#' empty_data <- data.frame()
#' validator_field_contents_blank(empty_data)
validator_field_contents_blank <- function(data) {
  return(TRUE)
}

#' Field types validator for csfmt_rts_data_v1 schema
#' 
#' Validates that field types conform to the csfmt_rts_data_v1 schema specification.
#' This validator ensures that the first 16 fields match the expected structure
#' for real-time surveillance data format version 1.
#' 
#' @param db_field_types A named character vector of database field types
#' @return TRUE if field types are valid for csfmt_rts_data_v1, FALSE otherwise
#' @export
#' @examples
#' # Valid field types for csfmt_rts_data_v1
#' valid_fields <- c(
#'   "granularity_time" = "TEXT",
#'   "granularity_geo" = "TEXT", 
#'   "country_iso3" = "TEXT",
#'   "location_code" = "TEXT",
#'   "border" = "INTEGER",
#'   "age" = "TEXT",
#'   "sex" = "TEXT",
#'   "isoyear" = "INTEGER",
#'   "isoweek" = "INTEGER",
#'   "isoyearweek" = "TEXT",
#'   "season" = "TEXT",
#'   "seasonweek" = "DOUBLE",
#'   "calyear" = "INTEGER",
#'   "calmonth" = "INTEGER",
#'   "calyearmonth" = "TEXT",
#'   "date" = "DATE",
#'   "cases_n" = "INTEGER"
#' )
#' validator_field_types_csfmt_rts_data_v1(valid_fields)
#' 
#' # Invalid field types (wrong structure)
#' invalid_fields <- c("id" = "INTEGER", "name" = "TEXT")
#' validator_field_types_csfmt_rts_data_v1(invalid_fields)
validator_field_types_csfmt_rts_data_v1 <- function(db_field_types) {
  if (!inherits(db_field_types, "character")) {
    return(FALSE)
  }
  if (!length(db_field_types) >= 16) {
    return(FALSE)
  }
  if (!identical(
    db_field_types[1:16],
    c(
      "granularity_time" = "TEXT",
      "granularity_geo" = "TEXT",
      "country_iso3" = "TEXT",
      "location_code" = "TEXT",
      "border" = "INTEGER",
      "age" = "TEXT",
      "sex" = "TEXT",
      "isoyear" = "INTEGER",
      "isoweek" = "INTEGER",
      "isoyearweek" = "TEXT",
      "season" = "TEXT",
      "seasonweek" = "DOUBLE",
      "calyear" = "INTEGER",
      "calmonth" = "INTEGER",
      "calyearmonth" = "TEXT",
      "date" = "DATE"
    )
  )) {
    return(FALSE)
  }

  return(TRUE)
}

#' Field contents validator for csfmt_rts_data_v1 schema
#' 
#' Validates that data contents conform to the csfmt_rts_data_v1 schema specification.
#' This validator checks that granularity_time and granularity_geo fields contain
#' valid values according to the surveillance data format requirements.
#' 
#' @param data A data.frame or data.table containing the data to validate
#' @return TRUE if data is valid for csfmt_rts_data_v1, FALSE otherwise (with error attribute)
#' @export
#' @examples
#' # Valid data for csfmt_rts_data_v1
#' valid_data <- data.frame(
#'   granularity_time = c("date", "isoyearweek", "total"),
#'   granularity_geo = c("nation", "county", "municip"),
#'   stringsAsFactors = FALSE
#' )
#' validator_field_contents_csfmt_rts_data_v1(valid_data)
#' 
#' # Invalid data (wrong granularity_geo value)
#' invalid_data <- data.frame(
#'   granularity_time = "date",
#'   granularity_geo = "invalid_geo",
#'   stringsAsFactors = FALSE
#' )
#' validator_field_contents_csfmt_rts_data_v1(invalid_data)
validator_field_contents_csfmt_rts_data_v1 <- function(data) {
  for (i in unique(data$granularity_time)) {
    if (sum(stringr::str_detect(
      i,
      c(
        "date",
        "isoyear",
        "isoyearweek",
        "^event",
        "total"
      )
    )) == 0) {
      retval <- FALSE
      attr(retval, "var") <- "granularity_time"
      return(retval)
    }
  }

  if (sum(!unique(data$granularity_geo) %in% c(
    "nation",
    "region",
    "hospitaldistrict",
    "county",
    "municip",
    "wardoslo",
    "extrawardoslo",
    "wardbergen",
    "wardtrondheim",
    "wardstavanger",
    "missingwardoslo",
    "missingwardbergen",
    "missingwardtrondheim",
    "missingwardstavanger",
    "ward",
    "station",
    "georegion",
    "baregion",
    "missingcounty",
    "missingmunicip",
    "notmainlandcounty",
    "notmainlandmunicip",
    "lab"
  )) > 0) {
    retval <- FALSE
    attr(retval, "var") <- "granularity_geo"
    return(retval)
  }

  if (sum(!unique(data$border) %in% c(
    "2020",
    "2024"
  )) > 0) {
    retval <- FALSE
    attr(retval, "var") <- "border"
    return(retval)
  }

  if (sum(!unique(data$sex) %in% c(
    "male",
    "female",
    "missing",
    "total"
  )) > 0) {
    retval <- FALSE
    attr(retval, "var") <- "sex"
    return(retval)
  }

  if (!inherits(data$date, "Date")) {
    retval <- FALSE
    attr(retval, "var") <- "date"
    return(retval)
  }

  return(TRUE)
}

#' Field types validator for csfmt_rts_data_v2 schema
#' 
#' Validates that field types conform to the csfmt_rts_data_v2 schema specification.
#' This validator ensures that the first 18 fields match the expected structure
#' for real-time surveillance data format version 2.
#' 
#' @param db_field_types A named character vector of database field types
#' @return TRUE if field types are valid for csfmt_rts_data_v2, FALSE otherwise
#' @export
#' @examples
#' # Valid field types for csfmt_rts_data_v2 (includes additional fields)
#' valid_fields_v2 <- c(
#'   "granularity_time" = "TEXT",
#'   "granularity_geo" = "TEXT", 
#'   "country_iso3" = "TEXT",
#'   "location_code" = "TEXT",
#'   "border" = "INTEGER",
#'   "age" = "TEXT",
#'   "sex" = "TEXT",
#'   "isoyear" = "INTEGER",
#'   "isoweek" = "INTEGER",
#'   "isoyearweek" = "TEXT",
#'   "season" = "TEXT",
#'   "seasonweek" = "DOUBLE",
#'   "calyear" = "INTEGER",
#'   "calmonth" = "INTEGER",
#'   "calyearmonth" = "TEXT",
#'   "date" = "DATE",
#'   "tag_outcome" = "TEXT",
#'   "tag_type" = "TEXT",
#'   "cases_n" = "INTEGER"
#' )
#' validator_field_types_csfmt_rts_data_v2(valid_fields_v2)
validator_field_types_csfmt_rts_data_v2 <- function(db_field_types) {
  if (!inherits(db_field_types, "character")) {
    return(FALSE)
  }
  if (!length(db_field_types) >= 18) {
    return(FALSE)
  }
  if (!identical(
    db_field_types[1:18],
    c(
      "granularity_time" = "TEXT",
      "granularity_geo" = "TEXT",
      "country_iso3" = "TEXT",
      "location_code" = "TEXT",
      "border" = "INTEGER",
      "age" = "TEXT",
      "sex" = "TEXT",
      "isoyear" = "INTEGER",
      "isoweek" = "INTEGER",
      "isoyearweek" = "TEXT",
      "isoquarter" = "INTEGER",
      "isoyearquarter" = "TEXT",
      "season" = "TEXT",
      "seasonweek" = "DOUBLE",
      "calyear" = "INTEGER",
      "calmonth" = "INTEGER",
      "calyearmonth" = "TEXT",
      "date" = "DATE"
    )
  )) {
    return(FALSE)
  }

  return(TRUE)
}

#' Field contents validator for csfmt_rts_data_v2 schema
#' 
#' Validates that data contents conform to the csfmt_rts_data_v2 schema specification.
#' This validator checks that granularity_time and granularity_geo fields contain
#' valid values according to the surveillance data format requirements for version 2.
#' 
#' @param data A data.frame or data.table containing the data to validate
#' @return TRUE if data is valid for csfmt_rts_data_v2, FALSE otherwise (with error attribute)
#' @export
#' @examples
#' # Valid data for csfmt_rts_data_v2
#' valid_data_v2 <- data.frame(
#'   granularity_time = c("date", "isoyearweek", "total"),
#'   granularity_geo = c("nation", "county", "municip"),
#'   stringsAsFactors = FALSE
#' )
#' validator_field_contents_csfmt_rts_data_v2(valid_data_v2)
validator_field_contents_csfmt_rts_data_v2 <- function(data) {
  for (i in unique(data$granularity_time)) {
    if (sum(stringr::str_detect(
      i,
      c(
        "date",
        "isoyear",
        "isoyearweek",
        "isoyearquarter",
        "season",
        "^event",
        "total"
      )
    )) == 0) {
      retval <- FALSE
      attr(retval, "var") <- "granularity_time"
      return(retval)
    }
  }

  if (sum(!unique(data$granularity_geo) %in% c(
    "nation",
    "georegion",
    "hospitaldistrict",
    "county",
    "municip",
    "wardoslo",
    "extrawardoslo",
    "wardbergen",
    "wardtrondheim",
    "wardstavanger",
    "missingwardoslo",
    "missingwardbergen",
    "missingwardtrondheim",
    "missingwardstavanger",
    "ward",
    "station",
    "baregion",
    "missingcounty",
    "missingmunicip",
    "notmainlandcounty",
    "notmainlandmunicip",
    "lab"
  )) > 0) {
    retval <- FALSE
    attr(retval, "var") <- "granularity_geo"
    return(retval)
  }

  if (sum(!unique(data$border) %in% c(
    "2020",
    "2024"
  )) > 0) {
    retval <- FALSE
    attr(retval, "var") <- "border"
    return(retval)
  }

  if (sum(!unique(data$sex) %in% c(
    "male",
    "female",
    "total"
  )) > 0) {
    retval <- FALSE
    attr(retval, "var") <- "sex"
    return(retval)
  }

  if (!inherits(data$date, "Date")) {
    retval <- FALSE
    attr(retval, "var") <- "date"
    return(retval)
  }

  return(TRUE)
}

# DBTable_v9 ----
#' R6 Class representing a database table with advanced data management capabilities
#'
#' @description
#' A comprehensive database table management class that provides high-level operations
#' for data manipulation, schema validation, and table administration. This class
#' combines database connectivity with data validation and efficient bulk operations.
#'
#' @details
#' The DBTable_v9 class is a sophisticated database table abstraction that provides:
#' 
#' \strong{Core functionality:}
#' \itemize{
#'   \item Table creation and schema management
#'   \item Data insertion with bulk loading capabilities
#'   \item Upsert operations (insert or update)
#'   \item Index management (creation, deletion)
#'   \item Data validation through customizable validators
#'   \item Integration with dplyr for data queries
#' }
#' 
#' \strong{Advanced features:}
#' \itemize{
#'   \item Automatic table creation based on field specifications
#'   \item Schema validation with custom validator functions
#'   \item Efficient bulk data loading using database-specific methods
#'   \item Index optimization for query performance
#'   \item Cross-database compatibility (SQL Server, PostgreSQL)
#' }
#' 
#' \strong{Data validation:}
#' The class supports custom validation functions for both field types and data contents,
#' ensuring data integrity and schema compliance.
#'
#' @import data.table
#' @import R6
#' @export DBTable_v9
#' @examples
#' \dontrun{
#' # Create database connection
#' db_config <- list(
#'   driver = "ODBC Driver 17 for SQL Server",
#'   server = "localhost",
#'   db = "mydb",
#'   user = "myuser",
#'   password = "mypass"
#' )
#' 
#' # Define table schema
#' field_types <- c(
#'   "id" = "INTEGER",
#'   "name" = "TEXT",
#'   "value" = "DOUBLE",
#'   "date_created" = "DATE"
#' )
#' 
#' # Create table object
#' my_table <- DBTable_v9$new(
#'   dbconfig = db_config,
#'   table_name = "my_data_table",
#'   field_types = field_types,
#'   keys = c("id"),
#'   validator_field_types = validator_field_types_blank,
#'   validator_field_contents = validator_field_contents_blank
#' )
#' 
#' # Create table in database
#' my_table$create_table()
#' 
#' # Insert data
#' sample_data <- data.frame(
#'   id = 1:3,
#'   name = c("Alice", "Bob", "Charlie"),
#'   value = c(10.5, 20.3, 15.7),
#'   date_created = as.Date("2023-01-01")
#' )
#' my_table$insert_data(sample_data)
#' 
#' # Query data using dplyr
#' result <- my_table$tbl() |>
#'   dplyr::filter(value > 15) |>
#'   dplyr::collect()
#' 
#' # Add indexes for performance
#' my_table$add_indexes(c("name", "date_created"))
#' 
#' # Upsert (insert or update) data
#' new_data <- data.frame(
#'   id = 2:4,
#'   name = c("Bob_Updated", "Charlie", "David"),
#'   value = c(25.0, 15.7, 30.2),
#'   date_created = as.Date("2023-01-02")
#' )
#' my_table$upsert_data(new_data)
#' }
DBTable_v9 <- R6::R6Class(
  "DBTable_v9",

  # public ----
  public = list(
    #' @field dbconnection Database connection.
    dbconnection = NULL,
    #' @field dbconfig Configuration details of the database.
    dbconfig = NULL,
    #' @field table_name Name of the table in the database.
    table_name = NULL,
    #' @field table_name_short_for_mssql_fully_specified_for_postgres Fully specified name of the table in the database (e.g. \[db\].\[dbo\].\[table_name\]).
    table_name_short_for_mssql_fully_specified_for_postgres = NULL,
    #' @field table_name_short_for_mssql_fully_specified_for_postgres_text Fully specified name of the table in the database (e.g. \[db\].\[dbo\].\[table_name\]).
    table_name_short_for_mssql_fully_specified_for_postgres_text = NULL,
    #' @field table_name_fully_specified Fully specified name of the table in the database (e.g. \[db\].\[dbo\].\[table_name\]).
    table_name_fully_specified = NULL,
    #' @field table_name_fully_specified_text Fully specified name of the table in the database (e.g. \[db\].\[dbo\].\[table_name\]) as a text string.
    table_name_fully_specified_text = NULL,
    #' @field field_types The types of each column in the database table (INTEGER, DOUBLE, TEXT, BOOLEAN, DATE, DATETIME).
    field_types = NULL,
    #' @field field_types_with_length The same as \code{field_types} but with \code{(100)} added to the end of all TEXT fields.
    field_types_with_length = NULL,
    #' @field keys The combination of variables that uniquely identify each row in the database.
    keys = NULL,
    #' @field keys_with_length The same as \code{keys} but with \code{(100)} added to the end of all TEXT fields.
    keys_with_length = NULL,
    #' @field indexes A named list of vectors (generally "ind1", "ind2", etc.) that improves the speed of data retrieval operations on a database table.
    indexes = NULL,
    #' @field validator_field_contents A function that validates the data before it is inserted into the database.
    validator_field_contents = NULL,
    #' @field load_folder A temporary folder that is used to write data to before inserting into the database.
    load_folder = tempdir(check = T),
    #' @field censors A named list of censors.
    censors = NULL,

    #' @description
    #' Create a new DBTable_v9 object.
    #'
    #' @param dbconfig Configuration details of the database (driver, server, port, db, schema, user, password, trusted_connection, sslmode, role_create_table).
    #' @param table_name Name of the table in the database.
    #' @param field_types The types of each column in the database table (INTEGER, DOUBLE, TEXT, BOOLEAN, DATE, DATETIME).
    #' @param keys The combination of these variables uniquely identifies each row of data in the table.
    #' @param indexes A named list of vectors (generally "ind1", "ind2", etc.) that improves the speed of data retrieval operations on a database table.
    #' @param validator_field_types A function that validates the \code{field_types} before the DB schema is created.
    #' @param validator_field_contents A function that validates the data before it is inserted into the database.
    #' @return A new `DBTable_v9` object.
    initialize = function(
    dbconfig,
    table_name,
    field_types,
    keys,
    indexes = NULL,
    validator_field_types = validator_field_types_blank,
    validator_field_contents = validator_field_contents_blank
    ) {

      force(dbconfig)
      self$dbconfig <- list()
      self$dbconfig$driver <- dbconfig$driver
      self$dbconfig$server <- dbconfig$server
      self$dbconfig$port <- dbconfig$port
      self$dbconfig$db <- dbconfig$db
      self$dbconfig$schema <- dbconfig$schema
      self$dbconfig$user <- dbconfig$user
      self$dbconfig$password <- dbconfig$password
      self$dbconfig$trusted_connection <- dbconfig$trusted_connection
      self$dbconfig$sslmode <- dbconfig$sslmode
      self$dbconfig$role_create_table <- dbconfig$role_create_table

      self$dbconnection <- DBConnection_v9$new(
        driver = self$dbconfig$driver,
        server = self$dbconfig$server,
        port = self$dbconfig$port,
        db = self$dbconfig$db,
        schema = self$dbconfig$schema,
        user = self$dbconfig$user,
        password = self$dbconfig$password,
        trusted_connection = self$dbconfig$trusted_connection,
        sslmode = self$dbconfig$sslmode,
        role_create_table = self$dbconfig$role_create_table
      )

      force(table_name)
      self$table_name <- table_name

      if(self$dbconfig$driver %in% c("ODBC Driver 17 for SQL Server")){
        table_fully_specified_vec = c(self$dbconfig$db, self$dbconfig$schema, self$table_name)
      } else {
        table_fully_specified_vec = c(self$dbconfig$schema, self$table_name)
      }
      self$table_name_fully_specified_text <- paste(table_fully_specified_vec, collapse = ".") |>
        stringr::str_remove_all("\\[]\\.")

      if(self$dbconfig$driver %in% c("ODBC Driver 17 for SQL Server")){
        self$table_name_fully_specified <- self$table_name_fully_specified_text
        self$table_name_short_for_mssql_fully_specified_for_postgres <- self$table_name
        self$table_name_short_for_mssql_fully_specified_for_postgres_text <- self$table_name
      } else {
        self$table_name_fully_specified <- DBI::Id(
          #database = self$dbconfig$db, this could be catalog??
          schema = self$dbconfig$schema,
          table = self$table_name
        )
        self$table_name_short_for_mssql_fully_specified_for_postgres <- self$table_name_fully_specified
        self$table_name_short_for_mssql_fully_specified_for_postgres_text <- self$table_name_fully_specified_text
      }

      force(field_types)
      self$field_types <- field_types
      self$field_types_with_length <- field_types

      force(keys)
      self$keys <- keys
      self$keys_with_length <- keys

      force(indexes)
      self$indexes <- indexes

      # validators
      if (!is.null(validator_field_types)) if (!validator_field_types(self$field_types)) stop(glue::glue("field_types not validated in {table_name}"))
      self$validator_field_contents <- validator_field_contents

      # db_field_types_with_lengths
      ind <- self$field_types == "TEXT"
      ind_text_with_specific_length <- stringr::str_detect(self$field_types, "TEXT")
      ind_text_with_specific_length[ind] <- FALSE
      if (sum(ind) > 0) {
        self$field_types_with_length[ind] <- paste0(self$field_types_with_length[ind], " (100)")
      }
      if (sum(ind_text_with_specific_length) > 0) {
        lengths <- stringr::str_extract(self$field_types[ind_text_with_specific_length], "\\([0-9]*\\)")
        self$field_types_with_length[ind_text_with_specific_length] <- paste0(self$field_types_with_length[ind_text_with_specific_length], " ", lengths)
      }

      # remove numbers from field_types
      naming <- names(self$field_types)
      self$field_types <- stringr::str_remove(self$field_types, " \\([0-9]*\\)")
      names(self$field_types) <- naming
      # fixing indexes
      self$keys_with_length <- self$field_types_with_length[self$keys]
    },

    #' @description
    #' Class-specific print function.
    #' @param ... Not in use.
    print = function(...) {
      if (!self$dbconnection$is_connected()) {
        if(requireNamespace("crayon", quietly = TRUE)) {
          cat(self$table_name_fully_specified_text, crayon::bgRed(crayon::white("(disconnected)\n\n")))
        } else {
          cat(self$table_name_fully_specified_text, "(disconnected)\n\n")
        }
      } else {
        if(requireNamespace("crayon", quietly = TRUE)) {
          cat(self$table_name_fully_specified_text, crayon::bgCyan(crayon::white("(connected)\n\n")))
        } else {
          cat(self$table_name_fully_specified_text, "(connected)\n\n")
        }
      }
      width_of_numbering <- nchar(length(self$field_types))
      for (i in seq_along(self$field_types)) {
        number <- formatC(i, width = width_of_numbering)
        x_name <- names(self$field_types)[i]
        x_type <- self$field_types[i]
        if(x_name %in% self$keys){
          x_key <- if(requireNamespace("crayon", quietly = TRUE)) crayon::bgRed(crayon::white("(KEY)")) else "(KEY)"
        } else {
          x_key <- ""
        }
        cat(" ", number, ": ", x_name, " (", x_type, ") ", x_key, "\n", sep = "")
      }
      cat("\n")

      invisible(self)
    },

    #' @description
    #' Connect from the database
    connect = function() {
      self$dbconnection$connect()
      private$lazy_creation_of_table()
    },

    #' @description
    #' Disconnect from the database
    disconnect = function() {
      self$dbconnection$disconnect()
    },

    #' @description
    #' Does the table exist
    table_exists = function() {
      return(DBI::dbExistsTable(self$dbconnection$autoconnection, self$table_name_short_for_mssql_fully_specified_for_postgres))
    },

    #' @description
    #' Create the database table
    create_table = function() {
      # self$connect calls self$create_table.
      # cannot have infinite loop
      create_tab <- TRUE
      if (self$table_exists()) {
        if (!private$check_fields_match()) {
          message(glue::glue("Dropping table {self$table_name} because fields dont match"))
          self$remove_table()
        } else {
          create_tab <- FALSE
        }
      }
      if (create_tab) {
        message(glue::glue("Creating table {self$table_name}"))
        create_table(
          connection = self$dbconnection$autoconnection,
          table = self$table_name_fully_specified,
          fields = self$field_types,
          keys = self$keys,
          role_create_table = self$dbconnection$config$role_create_table
        )
        private$add_constraint()
        self$add_indexes()
      }
    },

    #' @description
    #' Drop the database table
    remove_table = function() {
      if (self$table_exists()) {
        message(glue::glue("Dropping table {self$table_name}"))
        DBI::dbRemoveTable(self$dbconnection$autoconnection, self$table_name_short_for_mssql_fully_specified_for_postgres)
      }
    },

    #' @description
    #' Inserts data
    #' @param newdata The data to insert.
    #' @param confirm_insert_via_nrow Checks nrow() before insert and after insert. If nrow() has not increased sufficiently, then attempt an upsert.
    #' @param verbose Boolean.
    #' Inserts data into the database table
    insert_data = function(newdata, confirm_insert_via_nrow = FALSE, verbose = TRUE) {
      private$lazy_creation_of_table()
      if (is.null(newdata)) {
        return()
      }
      if (nrow(newdata) == 0) {
        return()
      }

      #newdata <- private$make_censored_data(newdata)

      validated <- self$validator_field_contents(newdata)
      if (!validated) stop(glue::glue("load_data_infile not validated in {self$table_name}. {attr(validated,'var')}"))

      # this will make the insert go faster, because
      # the data will be sorted
      # setkeyv(newdata, self$keys)
      infile <- random_file(private$load_folder_fn(), extra_insert = if(requireNamespace("digest", quietly = TRUE)) digest::digest(newdata[1,]) else "")
      load_data_infile(
        connection = self$dbconnection$autoconnection,
        dbconfig = self$dbconnection$config,
        table = self$table_name_short_for_mssql_fully_specified_for_postgres,
        dt = newdata,
        file = infile
      )

      if(confirm_insert_via_nrow){
        nrow_after <- self$nrow(use_count = TRUE)
        if(nrow_after < nrow(newdata)){
          message("After insert have ", nrow_after, " rows. Tried to insert ", nrow(newdata), ". Now trying upsert.")

          self$upsert_data(
            newdata = newdata,
            drop_indexes = NULL,
            verbose = verbose
          )
          nrow_after <- self$nrow(use_count = TRUE)
          if(nrow_after < nrow(newdata)){
            message("After upsert have ", nrow_after, " rows. Tried to upsert ", nrow(newdata), ".")
            stop("Upsert failed")
          }
        }
      }
    },

    #' @description
    #' Upserts data into the database table
    #' @param newdata The data to insert.
    #' @param drop_indexes A vector containing the indexes to be dropped before upserting (can increase performance).
    #' @param verbose Boolean.
    upsert_data = function(newdata, drop_indexes = names(self$indexes), verbose = TRUE) {
      private$lazy_creation_of_table()
      if (is.null(newdata)) {
        return()
      }
      if (nrow(newdata) == 0) {
        return()
      }

      # newdata <- private$make_censored_data(newdata)

      validated <- self$validator_field_contents(newdata)
      if (!validated) stop(glue::glue("upsert_load_data_infile not validated in {self$table_name}. {attr(validated,'var')}"))

      # this will make the insert go faster, because
      # the data will be sorted

      infile <- random_file(private$load_folder_fn(), extra_insert = if(requireNamespace("digest", quietly = TRUE)) digest::digest(newdata[1,]) else "")
      upsert_load_data_infile(
        connection = self$dbconnection$autoconnection,
        dbconfig = self$dbconnection$config,
        table = self$table_name_short_for_mssql_fully_specified_for_postgres,
        dt = newdata[, names(self$field_types), with = F],
        file = infile,
        fields = names(self$field_types),
        keys = self$keys,
        drop_indexes = drop_indexes
      )
    },

    #' @description
    #' Drops all rows in the database table
    drop_all_rows = function() {
      private$lazy_creation_of_table()
      drop_all_rows(connection = self$dbconnection$autoconnection, self$table_name_fully_specified_text)

    },

    #' @description
    #' Drops rows in the database table according to the SQL condition.
    #' @param condition SQL text condition.
    drop_rows_where = function(condition) {
      private$lazy_creation_of_table()

      drop_rows_where(
        connection = self$dbconnection$autoconnection,
        self$table_name_short_for_mssql_fully_specified_for_postgres_text,
        condition
      )
    },

    #' @description
    #' Keeps rows in the database table according to the SQL condition.
    #' @param condition SQL text condition.
    keep_rows_where = function(condition) {
      private$lazy_creation_of_table()
      keep_rows_where(connection = self$dbconnection$autoconnection, self$table_name_short_for_mssql_fully_specified_for_postgres_text, condition)
      private$add_constraint()
    },

    #' @description
    #' Drops all rows in the database table and then upserts data.
    #' @param newdata The data to insert.
    #' @param drop_indexes A vector containing the indexes to be dropped before upserting (can increase performance).
    #' @param verbose Boolean.
    drop_all_rows_and_then_upsert_data = function(newdata, drop_indexes = names(self$indexes), verbose = TRUE) {
      private$lazy_creation_of_table()
      self$drop_all_rows()
      self$upsert_data(
        newdata = newdata,
        drop_indexes = drop_indexes,
        verbose = verbose
      )
    },

    #' @description
    #' Drops all rows in the database table and then inserts data.
    #' @param newdata The data to insert.
    #' @param confirm_insert_via_nrow Checks nrow() before insert and after insert. If nrow() has not increased sufficiently, then attempt an upsert.
    #' @param verbose Boolean.
    drop_all_rows_and_then_insert_data = function(newdata, confirm_insert_via_nrow = FALSE, verbose = TRUE) {
      private$lazy_creation_of_table()
      self$drop_all_rows()
      self$insert_data(
        newdata = newdata,
        confirm_insert_via_nrow = confirm_insert_via_nrow,
        verbose = verbose
      )
    },

    #' @description
    #' Provides access to the database table via dplyr::tbl.
    tbl = function() {
      private$lazy_creation_of_table()
      retval <- self$dbconnection$autoconnection |>
        dplyr::tbl(self$table_name_short_for_mssql_fully_specified_for_postgres)

      return(retval)
    },

    #' @description
    #' Prints a template dplyr::select call that you can easily copy/paste for all your variables.
    print_dplyr_select = function() {
      private$lazy_creation_of_table()
      x <- self$tbl() |>
        head() |>
        dplyr::collect() |>
        names() |>
        paste0(., collapse = ",\n  ")
      x <- paste0("dplyr::select(\n  ", x, "\n) |>")
      cat(x)
    },

    #' @description
    #' Adds indexes to the database table from `self$indexes`
    add_indexes = function() {
      private$lazy_creation_of_table()
      for (i in names(self$indexes)) {
        message(glue::glue("Adding index {i}"))

        add_index(
          connection = self$dbconnection$autoconnection,
          table = self$table_name_short_for_mssql_fully_specified_for_postgres_text,
          index = i,
          keys = self$indexes[[i]]
        )
      }
    },

    #' @description
    #' Drops all indees from the database table
    drop_indexes = function() {
      private$lazy_creation_of_table()
      for (i in names(self$indexes)) {
        message(glue::glue("Dropping index {i}"))
        drop_index(
          connection = self$dbconnection$autoconnection,
          table = self$table_name_short_for_mssql_fully_specified_for_postgres_text,
          index = i
        )
      }
    },

    #' @description
    #' Confirms that the names and number of indexes in the database are the same as in the R code.
    #' Does not confirm the contents of the indexes!
    confirm_indexes = function() {
      indexes_db <- get_indexes(
        connection = self$dbconnection$autoconnection,
        table = self$table_name_short_for_mssql_fully_specified_for_postgres_text
      )
      indexes_self <- names(self$indexes)
      if(!identical(indexes_db, indexes_self)){
        self$drop_indexes()
        self$add_indexes()
      }
    },

    #' @description
    #' Gets the number of rows in the database table
    #' @param use_count If true, then uses the count command, which is slow but accurate. If false, then uses summary statistics, which is fast but inaccurate.
    nrow = function(use_count = FALSE){
      if(use_count){
        retval <- self$tbl() |>
          dplyr::summarize(n=n()) |>
          dplyr::collect()
        retval <- retval$n
      } else {
        retval <- get_table_names_and_info(self$dbconnection$autoconnection)
        retval <- retval[table_name %in% self$table_name]$nrow
      }
      return(retval)
    },

    #' @description
    #' Gets the information about the database table
    info = function(){
      retval <- get_table_names_and_info(self$dbconnection$autoconnection)
      retval <- retval[table_name %in% self$table_name]
      data.table::shouldPrint(retval)
      return(retval)
    }
  ),

  # private ----
  private = list(
    # Lazyload the creation of the table
    lazy_created_table = FALSE,
    lazy_creation_of_table = function(){
      if(!private$lazy_created_table){
        self$create_table()
        private$lazy_created_table <- TRUE
      }
    },

    check_fields_match = function() {
      fields <- DBI::dbListFields(self$dbconnection$autoconnection, self$table_name_short_for_mssql_fully_specified_for_postgres)
      retval <- identical(fields, names(self$field_types))
      if (retval == FALSE) {
        message(glue::glue(
          "given fields: {paste0(names(self$field_types),collapse=', ')}\n",
          "db fields: {paste0(fields,collapse=', ')}"
        ))
      }
      return(retval)
    },

    load_folder_fn = function() tempdir(check = T),

    add_constraint = function() {
      add_constraint(
        connection = self$dbconnection$autoconnection,
        table = self$table_name_short_for_mssql_fully_specified_for_postgres_text,
        keys = self$keys
      )
    },

    drop_constraint = function() {
      drop_constraint(
        connection = self$dbconnection$autoconnection,
        table = self$table_name_short_for_mssql_fully_specified_for_postgres_text
      )
    },

    make_censored_data = function(newdata) {
      d <- copy(newdata)
      for (i in seq_along(self$censors)) {
        self$censors[[i]](d)
      }
      return(d)
    },
    finalize = function() {
      # self$db_disconnect()
    }
  )
)

#' @import data.table ggplot2
#' @importFrom stats runif
#' @importFrom methods isClass getClass initialize
.onLoad <- function(libname, pkgname) {
  # Ensure methods package is available for S7 operations
  if (!requireNamespace("methods", quietly = TRUE)) {
    return(NULL)
  }

  # Register S7 methods for database utilities
  tryCatch({
    S7::methods_register()
  }, error = function(e) {
    # Silently continue if S7 registration fails
    NULL
  })

  # Try to register S4 classes and update method dispatch when package loads
  # This will re-run the registration in case odbc package is now available
  tryCatch({
    register_s4_classes()

    # Re-register methods with updated class information
    S7::methods_register()
  }, error = function(e) {
    # Silently continue if registration fails
    NULL
  })
}

.onAttach <- function(libname, pkgname) {
    version <- tryCatch(
      utils::packageDescription("csdb", fields = "Version"),
      warning = function(w){
        1
      }
    )

  packageStartupMessage(paste0(
    "csdb ",
    version,
    "\n",
    "https://niphr.github.io/csdb/"
  ))
}

#' Get table names, number of rows, and size information
#' 
#' Retrieves comprehensive information about database tables including their names,
#' row counts, and storage size metrics. This function provides database-specific
#' implementations for different database systems.
#' 
#' @param connection A database connection object (e.g., from \code{\link[DBI]{dbConnect}})
#' @return A data.table containing table information with columns:
#' \describe{
#'   \item{table_name}{Character. Name of the table}
#'   \item{nrow}{Numeric. Number of rows in the table}
#'   \item{size_total_gb}{Numeric. Total size of the table in gigabytes}
#'   \item{size_data_gb}{Numeric. Size of data in gigabytes}
#'   \item{size_index_gb}{Numeric. Size of indexes in gigabytes}
#' }
#' @export
#' @examples
#' \dontrun{
#' # Microsoft SQL Server example
#' con <- DBI::dbConnect(odbc::odbc(), 
#'                       driver = "ODBC Driver 17 for SQL Server",
#'                       server = "localhost", 
#'                       database = "mydb")
#' table_info <- get_table_names_and_info(con)
#' print(table_info)
#' DBI::dbDisconnect(con)
#' 
#' # PostgreSQL example  
#' con <- DBI::dbConnect(RPostgres::Postgres(),
#'                       host = "localhost",
#'                       dbname = "mydb",
#'                       user = "user")
#' table_info <- get_table_names_and_info(con)
#' print(table_info)
#' DBI::dbDisconnect(con)
#' }
#' @export
get_table_names_and_info <- function(connection) UseMethod("get_table_names_and_info")

#' @export
`get_table_names_and_info.Microsoft SQL Server` <- function(connection) {
  # Declare variables to avoid R CMD check NOTEs
  . <- NULL
  table_name <- NULL
  name <- NULL
  size_total_gb <- NULL
  reserved <- NULL
  size_data_gb <- NULL
  data <- NULL
  size_index_gb <- NULL
  index_size <- NULL
  rows <- NULL
  nrow <- NULL
  # table_rows <- connection %>%
  #   DBI::dbGetQuery("select o.name as table_name, i.rowcnt as n from sys.objects o join sys.sysindexes i on o.object_id = i.id where o.is_ms_shipped = 0 and i.rowcnt > 0 order by o.name") %>%
  #   setDT() %>% unique()

  # update stats
  # connection %>% DBI::dbExecute("sp_updatestats")
  # get the stats
  table_rows <- connection |>
    DBI::dbGetQuery("sp_msforeachtable 'sp_spaceused [?]'") |>
    setDT()
  table_rows[, size_total_gb := round(as.numeric(stringr::str_extract_all(reserved, "[0-9]+"))/1024/1024, digits = 2)]
  table_rows[, size_data_gb := round(as.numeric(stringr::str_extract_all(data, "[0-9]+"))/1024/1024, digits = 2)]
  table_rows[, size_index_gb := round(as.numeric(stringr::str_extract_all(index_size, "[0-9]+"))/1024/1024, digits = 2)]
  table_rows[, nrow := as.numeric(stringr::str_extract_all(rows, "[0-9]+"))]

  table_rows <- table_rows[,.(
    table_name = name,
    nrow,
    size_total_gb,
    size_data_gb,
    size_index_gb
  )]
  setorder(table_rows, table_name)

  data.table::shouldPrint(table_rows)
  return(table_rows)
}

#' @export
get_table_names_and_info.PostgreSQL <- function(connection) {

  sql = "SELECT table_name
     , row_estimate AS nrow
     , cast(total_bytes as decimal)/1073741824 AS size_total_gb
     , cast(data_bytes as decimal)/1073741824 AS size_data_gb
     , cast(index_bytes as decimal)/1073741824 AS size_index_gb
  FROM (
     SELECT *,
         total_bytes-index_bytes AS data_bytes
     FROM (
         SELECT nspname AS table_schema
                , relname AS table_name
                , c.reltuples AS row_estimate
                , pg_total_relation_size(c.oid) AS total_bytes
                , pg_indexes_size(c.oid) AS index_bytes
            FROM pg_class c
            LEFT JOIN pg_namespace n ON n.oid = c.relnamespace
            WHERE relkind = 'r'
     ) raw_storage
  ) storage_with_data_size
  order by table_name;"

  table_rows = DBI::dbGetQuery(connection, sql) |> setDT()

  data.table::shouldPrint(table_rows)
  return(table_rows)
}

# Database Utilities for csdb package
# This file contains S7 database methods and database-specific utility functions

# Database utility functions

#' Write data.table to file for database bulk insert
#'
#' Internal function to write data.table to CSV file with proper formatting
#' for database bulk insert operations. Handles special cases like infinites,
#' NaNs, and POSIXt objects.
#'
#' @param dt data.table to write
#' @param file Output file path
#' @param colnames Logical indicating whether to include column names
#' @param eol End of line character
#' @param quote Quoting behavior
#' @param na String to use for NA values
#' @param sep Column separator
#' @return NULL (called for side effects)
#' @keywords internal
#' @noRd
write_data_infile <- function(
    dt,
    file = paste0(tempfile(), ".csv"),
    colnames = T,
    eol = "\n",
    quote = "auto",
    na = "\\N",
    sep = ","
) {
  # infinites and NANs get written as text
  # which destroys the upload
  # we need to set them to NA
  for (i in names(dt)) {
    dt[is.infinite(get(i)), (i) := NA]
    dt[is.nan(get(i)), (i) := NA]
    if (inherits(dt[[i]], "POSIXt")) dt[, (i) := as.character(get(i))]
  }
  fwrite(dt,
         file = file,
         logical01 = T,
         na = na,
         col.names = colnames,
         eol = eol,
         quote = quote,
         sep = sep
  )
}

#' Truncate all rows from a database table
#'
#' Removes all rows from a database table using TRUNCATE TABLE command.
#' This is more efficient than DELETE for removing all rows.
#'
#' @param connection Database connection object
#' @param table Name of the table to truncate
#' @return NULL (called for side effects)
#' @keywords internal
#' @noRd
drop_all_rows <- function(connection, table) {
  a <- DBI::dbExecute(connection, glue::glue({
    "TRUNCATE TABLE {table};"
  }))
}

#' List indexes for a database table
#'
#' Internal function to list all indexes for a specific table.
#' Currently only supports Microsoft SQL Server.
#'
#' @param connection Database connection object
#' @param table Name of the table to list indexes for
#' @return data.frame with index information
#' @keywords internal
#' @noRd
list_indexes <- function(connection, table) {
  retval <- DBI::dbGetQuery(
    connection,
    glue::glue("select * from sys.indexes where object_id = (select object_id from sys.objects where name = '{table}')")
  )
  return(retval)
}

# S7 classes for database connections - register S4 classes for S7 dispatch
# This allows S7 to dispatch on actual S4 connection objects  
# We need to register the S4 classes with S7 first, then create methods for them

# Helper function to safely register S4 classes for S7 dispatch
register_s4_classes <- function() {
  # Register S4 classes from odbc package if available
  tryCatch({
    # Ensure odbc package is available before trying to register classes
    if (requireNamespace("odbc", quietly = TRUE)) {
      # Check if PostgreSQL S4 class exists in odbc package and register it
      if (methods::isClass("PostgreSQL") && 
          methods::getClass("PostgreSQL")@package == "odbc") {
        S7::S4_register(methods::getClass("PostgreSQL"))
      }
      
      # Check if Microsoft SQL Server S4 class exists and register it
      if (methods::isClass("Microsoft SQL Server") && 
          methods::getClass("Microsoft SQL Server")@package == "odbc") {
        S7::S4_register(methods::getClass("Microsoft SQL Server"))
      }
    }
  }, error = function(e) {
    # Silently continue if odbc package not available
    NULL
  })
  
  # Register DBI classes if available
  tryCatch({
    if (requireNamespace("DBI", quietly = TRUE)) {
      if (methods::isClass("DBIConnection")) {
        S7::S4_register(methods::getClass("DBIConnection"))
      }
    }
  }, error = function(e) NULL)
}

# Helper function to get or create database class objects for S7 dispatch
get_db_classes <- function() {
  # Try to get actual S4 classes first, fall back to S3 wrappers
  
  # PostgreSQL class
  db_postgres <- tryCatch({
    if (requireNamespace("odbc", quietly = TRUE) && 
        methods::isClass("PostgreSQL") &&
        methods::getClass("PostgreSQL")@package == "odbc") {
      methods::getClass("PostgreSQL")
    } else {
      S7::new_S3_class("PostgreSQL")
    }
  }, error = function(e) S7::new_S3_class("PostgreSQL"))
  
  # Microsoft SQL Server class
  db_mssql <- tryCatch({
    if (requireNamespace("odbc", quietly = TRUE) && 
        methods::isClass("Microsoft SQL Server") &&
        methods::getClass("Microsoft SQL Server")@package == "odbc") {
      methods::getClass("Microsoft SQL Server")
    } else {
      S7::new_S3_class("Microsoft SQL Server")
    }
  }, error = function(e) S7::new_S3_class("Microsoft SQL Server"))
  
  # Default/DBI class
  db_default <- tryCatch({
    if (requireNamespace("DBI", quietly = TRUE) && 
        methods::isClass("DBIConnection")) {
      methods::getClass("DBIConnection")
    } else {
      S7::new_S3_class("DBIConnection")
    }
  }, error = function(e) S7::new_S3_class("DBIConnection"))
  
  return(list(
    postgres = db_postgres,
    mssql = db_mssql,
    default = db_default
  ))
}

# Public function to re-register database methods
# This can be called if database connections are not working properly
#' Re-register database S7 methods
#' 
#' Re-registers S7 methods for database operations. Call this function if you 
#' encounter method dispatch errors with database connections.
#' 
#' @return NULL (called for side effects)
#' @keywords internal
#' @noRd
refresh_database_methods <- function() {
  # Re-register S4 classes
  register_s4_classes()
  
  # Re-get database class objects
  db_classes <- get_db_classes()
  db_postgres <<- db_classes$postgres
  db_mssql <<- db_classes$mssql
  db_default <<- db_classes$default
  
  # Re-register S7 methods
  S7::methods_register()
  
  message("Database methods refreshed. S4 classes re-registered and S7 methods updated.")
  invisible(NULL)
}

# Debug function to show method dispatch information
#' Debug database method dispatch
#' 
#' Shows information about registered classes and methods for debugging
#' method dispatch issues.
#' 
#' @param connection Optional database connection object to check
#' @return List with debugging information
#' @keywords internal
#' @noRd
debug_database_methods <- function(connection = NULL) {
  info <- list()
  
  # Check available packages
  info$packages <- list(
    odbc_available = requireNamespace("odbc", quietly = TRUE),
    DBI_available = requireNamespace("DBI", quietly = TRUE),
    S7_available = requireNamespace("S7", quietly = TRUE)
  )
  
  # Check registered S4 classes
  info$s4_classes <- list()
  if (info$packages$odbc_available) {
    info$s4_classes$PostgreSQL <- tryCatch({
      if (methods::isClass("PostgreSQL")) {
        list(
          exists = TRUE,
          package = methods::getClass("PostgreSQL")@package
        )
      } else {
        list(exists = FALSE)
      }
    }, error = function(e) list(error = e$message))
    
    info$s4_classes$MicrosoftSQLServer <- tryCatch({
      if (methods::isClass("Microsoft SQL Server")) {
        list(
          exists = TRUE,
          package = methods::getClass("Microsoft SQL Server")@package
        )
      } else {
        list(exists = FALSE)
      }
    }, error = function(e) list(error = e$message))
  }
  
  # Check connection object if provided
  if (!is.null(connection)) {
    info$connection <- list(
      class = class(connection),
      is_s4 = methods::is(connection, "S4"),
      package = attr(class(connection), "package")
    )
  }
  
  # Check current database class objects
  info$db_classes <- list(
    postgres_type = class(db_postgres),
    mssql_type = class(db_mssql),
    default_type = class(db_default)
  )
  
  return(info)
}

# Register S4 classes and get database class objects
register_s4_classes()
db_classes <- get_db_classes()
db_postgres <- db_classes$postgres
db_mssql <- db_classes$mssql
db_default <- db_classes$default

# S7 generic definitions (internal use only)
load_data_infile <- S7::new_generic("load_data_infile", "connection")
upsert_load_data_infile <- S7::new_generic("upsert_load_data_infile", "connection")
create_table <- S7::new_generic("create_table", "connection")
add_constraint <- S7::new_generic("add_constraint", "connection")
drop_constraint <- S7::new_generic("drop_constraint", "connection")
get_indexes <- S7::new_generic("get_indexes", "connection")
drop_index <- S7::new_generic("drop_index", "connection")
add_index <- S7::new_generic("add_index", "connection")
drop_rows_where <- S7::new_generic("drop_rows_where", "connection")
keep_rows_where <- S7::new_generic("keep_rows_where", "connection")
drop_table <- S7::new_generic("drop_table", "connection")

# S7 method definitions
# load_data_infile methods
S7::method(load_data_infile, db_default) <- function(connection, 
                                                       dbconfig = NULL, 
                                                       table, 
                                                       dt = NULL, 
                                                       file = "/xtmp/x123.csv", 
                                                       force_tablock = FALSE) {
  if (is.null(dt)) {
    return()
  }
  if (nrow(dt) == 0) {
    return()
  }

  t0 <- Sys.time()

  correct_order <- DBI::dbListFields(connection, table)
  if (length(correct_order) > 0) dt <- dt[, correct_order, with = F]
  write_data_infile(dt = dt, file = file)
  on.exit(unlink(file), add = T)

  sep <- ","
  eol <- "\n"
  quote <- '"'
  skip <- 0
  header <- T
  path <- normalizePath(file, winslash = "/", mustWork = TRUE)

  sql <- paste0(
    "LOAD DATA INFILE ", DBI::dbQuoteString(connection, path), "\n",
    "INTO TABLE ", DBI::dbQuoteIdentifier(connection, table), "\n",
    "CHARACTER SET utf8", "\n",
    "FIELDS TERMINATED BY ", DBI::dbQuoteString(connection, sep), "\n",
    "OPTIONALLY ENCLOSED BY ", DBI::dbQuoteString(connection, quote), "\n",
    "LINES TERMINATED BY ", DBI::dbQuoteString(connection, eol), "\n",
    "IGNORE ", skip + as.integer(header), " LINES \n",
    "(", paste0(correct_order, collapse = ","), ")"
  )
  DBI::dbExecute(connection, sql)

  t1 <- Sys.time()
  dif <- round(as.numeric(difftime(t1, t0, units = "secs")), 1)

  invisible()
}

S7::method(load_data_infile, db_mssql) <- function(connection, 
                                                     dbconfig = NULL, 
                                                     table, 
                                                     dt, 
                                                     file = tempfile(), 
                                                     force_tablock = FALSE) {
  if (is.null(dt)) {
    return()
  }
  if (nrow(dt) == 0) {
    return()
  }

  a <- Sys.time()

  correct_order <- DBI::dbListFields(connection, table)
  if (length(correct_order) > 0) dt <- dt[, correct_order, with = F]
  write_data_infile(
    dt = dt,
    file = file,
    colnames = F,
    eol = "\n",
    quote = FALSE,
    na = "",
    sep = "\t"
  )
  on.exit(unlink(file), add = T)

  format_file <- tempfile(tmpdir = tempdir(check = TRUE))
  on.exit(unlink(format_file), add = T)

  args <- c(
    table,
    "format",
    "nul",
    "-q",
    "-c",
    "-f",
    format_file,
    "-S",
    dbconfig$server,
    "-d",
    dbconfig$db,
    "-U",
    dbconfig$user,
    "-P",
    dbconfig$password
  )
  if (dbconfig$trusted_connection == "yes") {
    args <- c(args, "-T")
  }
  
  if (Sys.which("bcp") == "") {
    stop("bcp command not found. Please install SQL Server command line tools.")
  }
  
  system2(
    "bcp",
    args = args,
    stdout = NULL
  )

  if(FALSE){
    hint_arg <- NULL
  } else {
    hint_arg <- NULL
  }

  if (!is.null(key(dt))) {
    hint_arg <- c(hint_arg, paste0("ORDER(", paste0(key(dt), " ASC", collapse = ", "), ")"))
  }
  if (length(hint_arg) > 0) {
    hint_arg <- paste0(hint_arg, collapse = ", ")
    hint_arg <- paste0("-h '", hint_arg, "'")
  }

  args <- c(
    table,
    "in",
    file,
    "-a 16384",
    hint_arg,
    "-S",
    dbconfig$server,
    "-d",
    dbconfig$db,
    "-U",
    dbconfig$user,
    "-P",
    dbconfig$password,
    "-f",
    format_file,
    "-m",
    0
  )
  if (dbconfig$trusted_connection == "yes") {
    args <- c(args, "-T")
  }
  
  if (Sys.which("bcp") == "") {
    stop("bcp command not found. Please install SQL Server command line tools.")
  }
  
  system2(
    "bcp",
    args = args,
    stdout = NULL
  )

  b <- Sys.time()
  dif <- round(as.numeric(difftime(b, a, units = "secs")), 1)

  invisible()
}

S7::method(load_data_infile, db_postgres) <- function(connection, 
                                                        dbconfig = NULL, 
                                                        table, 
                                                        dt, 
                                                        file = tempfile(), 
                                                        force_tablock = FALSE) {
  if (is.null(dt)) {
    return()
  }
  if (nrow(dt) == 0) {
    return()
  }

  a <- Sys.time()

  table_text <- DBI::dbQuoteIdentifier(connection, table)

  correct_order <- DBI::dbListFields(connection, table)

  if (length(correct_order) > 0) {
    dt <- dt[, correct_order, with = F]
  }

  write_data_infile(
    dt = dt,
    file = file,
    colnames = F,
    eol = "\n",
    quote = FALSE,
    na = "",
    sep = "\t"
  )

  on.exit(unlink(file), add = T)

  sql <- sprintf(
    "\"\\copy %s (%s) from '%s' (FORMAT CSV, DELIMITER '\t')\"",
    table_text,
    paste(correct_order, collapse = ","),
    file
  )

  uri <- sprintf(
    "postgresql://%s:%s@%s:%s/%s",
    dbconfig$user,
    dbconfig$password,
    dbconfig$server,
    dbconfig$port,
    dbconfig$db
  )

  args <- c(
    "-U",
    dbconfig$user,
    "-c",
    sql,
    uri
  )

  if (Sys.which("psql") == "") {
    stop("psql command not found. Please install PostgreSQL command line tools.")
  }

  system2(
    "psql",
    args = args,
    stdout = FALSE
  )

  b <- Sys.time()
  dif <- round(as.numeric(difftime(b, a, units = "secs")), 1)

  invisible()
}

# Continue with upsert_load_data_infile methods
S7::method(upsert_load_data_infile, db_default) <- function(connection, 
                                                              dbconfig = NULL, 
                                                              table, 
                                                              dt, 
                                                              file = "/tmp/x123.csv", 
                                                              fields, 
                                                              keys = NULL, 
                                                              drop_indexes = NULL) {
  temp_name <- random_uuid()
  on.exit(DBI::dbRemoveTable(connection, temp_name), add = TRUE, after = FALSE)

  sql <- glue::glue("CREATE TEMPORARY TABLE {temp_name} LIKE {table};")
  DBI::dbExecute(connection, sql)

  if (!is.null(drop_indexes)) {
    for (i in drop_indexes) {
      try(
        DBI::dbExecute(
          connection,
          glue::glue("ALTER TABLE `{temp_name}` DROP INDEX `{i}`")
        ),
        TRUE
      )
    }
  }

  load_data_infile(
    connection = connection,
    dbconfig = dbconfig,
    table = temp_name,
    dt = dt,
    file = file
  )

  t0 <- Sys.time()

  vals_fields <- glue::glue_collapse(fields, sep = ", ")
  vals <- glue::glue("{fields} = VALUES({fields})")
  vals <- glue::glue_collapse(vals, sep = ", ")

  sql <- glue::glue("
    INSERT INTO {table} SELECT {vals_fields} FROM {temp_name}
    ON DUPLICATE KEY UPDATE {vals};
    ")
  DBI::dbExecute(connection, sql)

  t1 <- Sys.time()
  dif <- round(as.numeric(difftime(t1, t0, units = "secs")), 1)

  invisible()
}

S7::method(upsert_load_data_infile, db_mssql) <- function(connection, 
                                                            dbconfig, 
                                                            table, 
                                                            dt, 
                                                            file = tempfile(), 
                                                            fields, 
                                                            keys, 
                                                            drop_indexes = NULL) {
  temp_name <- paste0("tmp", random_uuid())
  on.exit(DBI::dbRemoveTable(connection, temp_name), add = TRUE, after = FALSE)

  sql <- glue::glue("SELECT * INTO {temp_name} FROM {table} WHERE 1 = 0;")
  DBI::dbExecute(connection, sql)

  load_data_infile(
    connection = connection,
    dbconfig = dbconfig,
    table = temp_name,
    dt = dt,
    file = file,
    force_tablock = TRUE
  )

  a <- Sys.time()
  add_index(
    connection = connection,
    table = temp_name,
    keys = keys
  )

  vals_fields <- glue::glue_collapse(fields, sep = ", ")
  vals <- glue::glue("{fields} = VALUES({fields})")
  vals <- glue::glue_collapse(vals, sep = ", ")

  sql_on_keys <- glue::glue("{t} = {s}", t = paste0("t.", keys), s = paste0("s.", keys))
  sql_on_keys <- paste0(sql_on_keys, collapse = " and ")

  sql_update_set <- glue::glue("{t} = {s}", t = paste0("t.", fields), s = paste0("s.", fields))
  sql_update_set <- paste0(sql_update_set, collapse = ", ")

  sql_insert_fields <- paste0(fields, collapse = ", ")
  sql_insert_s_fields <- paste0(paste0("s.", fields), collapse = ", ")

  sql <- glue::glue("
  MERGE {table} t
  USING {temp_name} s
  ON ({sql_on_keys})
  WHEN MATCHED
  THEN UPDATE SET
    {sql_update_set}
  WHEN NOT MATCHED BY TARGET
  THEN INSERT ({sql_insert_fields})
    VALUES ({sql_insert_s_fields});
  ")

  DBI::dbExecute(connection, sql)

  b <- Sys.time()
  dif <- round(as.numeric(difftime(b, a, units = "secs")), 1)

  invisible()
}

S7::method(upsert_load_data_infile, db_postgres) <- function(connection, 
                                                               dbconfig, 
                                                               table, 
                                                               dt, 
                                                               file = tempfile(), 
                                                               fields, 
                                                               keys, 
                                                               drop_indexes = NULL) {
  temp_name <- DBI::Id(schema = table@name[["schema"]], paste0("tmp", random_uuid()))
  temp_name_text <- DBI::dbQuoteIdentifier(connection, temp_name)
  table_text <- DBI::dbQuoteIdentifier(connection, table)

  on.exit(DBI::dbRemoveTable(connection, temp_name), add = TRUE, after = FALSE)

  sql <- glue::glue("SELECT * INTO {temp_name_text} FROM {table_text} WHERE 1 = 0;")
  DBI::dbExecute(connection, sql)

  load_data_infile(
    connection = connection,
    dbconfig = dbconfig,
    table = temp_name,
    dt = dt,
    file = file,
    force_tablock = TRUE
  )

  a <- Sys.time()
  add_index(
    connection = connection,
    table = temp_name,
    keys = keys,
    index = "ind" + random_uuid()
  )

  vals_fields <- glue::glue_collapse(fields, sep = ", ")
  vals <- glue::glue("{fields} = VALUES({fields})")
  vals <- glue::glue_collapse(vals, sep = ", ")

  sql_on_keys <- glue::glue("{t} = {s}", t = paste0("t.", keys), s = paste0("s.", keys))
  sql_on_keys <- paste0(sql_on_keys, collapse = " and ")

  update_fields <- setdiff(fields, keys)
  sql_update_set <- glue::glue("{t} = {s}", t = update_fields, s = paste0("s.", update_fields))
  sql_update_set <- paste0(sql_update_set, collapse = ", ")

  sql_insert_fields <- paste0(fields, collapse = ", ")
  sql_insert_s_fields <- paste0(paste0("s.", fields), collapse = ", ")

  sql <- glue::glue("
  MERGE INTO {table_text} t
  USING {temp_name_text} s
  ON ({sql_on_keys})
  WHEN MATCHED
  THEN UPDATE SET
    {sql_update_set}
  WHEN NOT MATCHED
  THEN INSERT ({sql_insert_fields})
    VALUES ({sql_insert_s_fields});
  ")

  DBI::dbExecute(connection, sql)

  b <- Sys.time()
  dif <- round(as.numeric(difftime(b, a, units = "secs")), 1)

  invisible()
}

# create_table methods
S7::method(create_table, db_default) <- function(connection, table, fields, keys = NULL, role_create_table = NULL, ...) {
  fields_new <- fields
  fields_new[fields == "TEXT"] <- "TEXT CHARACTER SET utf8 COLLATE utf8_unicode_ci"

  sql <- DBI::sqlCreateTable(connection, table, fields_new,
                             row.names = F, temporary = F
  )
  DBI::dbExecute(connection, sql)
}

S7::method(create_table, db_mssql) <- function(connection, table, fields, keys = NULL, role_create_table = NULL, ...) {
  fields_new <- fields
  fields_new[fields == "TEXT"] <- "NVARCHAR (1000)"
  fields_new[fields == "DOUBLE"] <- "FLOAT"
  fields_new[fields == "BOOLEAN"] <- "BIT"

  if (!is.null(keys)) fields_new[names(fields_new) %in% keys] <- paste0(fields_new[names(fields_new) %in% keys], " NOT NULL")

  sql <- DBI::sqlCreateTable(
    connection,
    table,
    fields_new,
    row.names = F,
    temporary = F
  ) |>
    stringr::str_replace("\\\\", "\\") |>
    stringr::str_replace("\"", "") |>
    stringr::str_replace("\"", "")
  DBI::dbExecute(connection, sql)
}

S7::method(create_table, db_postgres) <- function(connection, table, fields, keys = NULL, role_create_table = NULL, ...) {
  fields_new <- fields
  fields_new[fields == "TEXT"] <- "VARCHAR"
  fields_new[fields == "DOUBLE"] <- "REAL"
  fields_new[fields == "BOOLEAN"] <- "BIT"
  fields_new[fields == "DATETIME"] <- "TIMESTAMP"

  if (!is.null(keys)) fields_new[names(fields_new) %in% keys] <- paste0(fields_new[names(fields_new) %in% keys], " NOT NULL")

  sql <- DBI::sqlCreateTable(
    connection,
    table,
    fields_new,
    row.names = F,
    temporary = F
  ) |>
    stringr::str_replace("\\\\", "\\") |>
    stringr::str_replace("\"", "") |>
    stringr::str_replace("\"", "")

  if(!is.na(role_create_table)) if(role_create_table!="x"){
    sql <- paste0("SET ROLE ", DBI::dbQuoteIdentifier(connection, role_create_table), "; ", sql,"; RESET ROLE")
  }

  DBI::dbExecute(connection, sql)
}

# add_constraint methods
S7::method(add_constraint, db_default) <- function(connection, table, keys) {
  t0 <- Sys.time()

  primary_keys <- glue::glue_collapse(keys, sep = ", ")
  constraint <- glue::glue("PK_{table}") |>
    stringr::str_remove_all("\\.") |>
    stringr::str_remove_all("\\[") |>
    stringr::str_remove_all("]")
  sql <- glue::glue("
          ALTER table {table}
          ADD CONSTRAINT {constraint} PRIMARY KEY CLUSTERED ({primary_keys});")
  a <- DBI::dbExecute(connection, sql)
  t1 <- Sys.time()
  dif <- round(as.numeric(difftime(t1, t0, units = "secs")), 1)
}

S7::method(add_constraint, db_postgres) <- function(connection, table, keys) {
  t0 <- Sys.time()

  primary_keys <- glue::glue_collapse(keys, sep = ", ")
  constraint <- glue::glue("PK_{table}") |>
    stringr::str_remove_all("\\.") |>
    stringr::str_remove_all("\\[") |>
    stringr::str_remove_all("]")
  sql <- glue::glue(
    "ALTER table {table}
    ADD CONSTRAINT {constraint}
    PRIMARY KEY ({primary_keys});"
  )

  a <- DBI::dbExecute(connection, sql)

  t1 <- Sys.time()
  dif <- round(as.numeric(difftime(t1, t0, units = "secs")), 1)
}

# drop_constraint methods
S7::method(drop_constraint, db_default) <- function(connection, table) {
  constraint <- glue::glue("PK_{table}") |>
    stringr::str_remove_all("\\.") |>
    stringr::str_remove_all("\\[") |>
    stringr::str_remove_all("]")
  sql <- glue::glue("
          ALTER table {table}
          DROP CONSTRAINT {constraint};")
  try(a <- DBI::dbExecute(connection, sql), TRUE)
}

# get_indexes methods
S7::method(get_indexes, db_mssql) <- function(connection, table){
  index_name <- NULL
  table_name <- NULL

  table_rows <- connection |>
    DBI::dbGetQuery("select o.name as table_name, i.name as index_name from sys.objects o join sys.sysindexes i on o.object_id = i.id where o.is_ms_shipped = 0 and i.rowcnt > 0 order by o.name") |>
    dplyr::filter(!is.na(index_name) & !stringr::str_detect(index_name, "^PK")) |>
    setDT()
  retval <- table_rows[table_name %in% table]$index_name
  return(retval)
}

S7::method(get_indexes, db_postgres) <- function(connection, table){
  index_name <- NULL
  table_name <- NULL

  sql <- "
    select tablename, indexname
    from pg_indexes
  "

  table_rows <- connection |>
    DBI::dbGetQuery(sql) |>
    dplyr::filter(!is.na(indexname) & !stringr::str_detect(indexname, "^pk")) |>
    setDT()
  retval <- table_rows[tablename %in% table]$indexname
  return(retval)
}

# drop_index methods
S7::method(drop_index, db_default) <- function(connection, table, index) {
  try(
    DBI::dbExecute(
      connection,
      glue::glue("ALTER TABLE `{table}` DROP INDEX `{index}`")
    ),
    TRUE
  )
}

S7::method(drop_index, db_mssql) <- function(connection, table, index) {
  try(
    DBI::dbExecute(
      connection,
      glue::glue("DROP INDEX {table}.{index}")
    ),
    TRUE
  )
}

S7::method(drop_index, db_postgres) <- function(connection, table, index) {
  try(
    DBI::dbExecute(
      connection,
      glue::glue("DROP INDEX IF EXISTS {index}")
    ),
    TRUE
  )
}

# add_index methods
S7::method(add_index, db_default) <- function(connection, table, index, keys) {
  keys <- glue::glue_collapse(keys, sep = ", ")

  sql <- glue::glue("
    ALTER TABLE `{table}` ADD INDEX `{index}` ({keys})
    ;")
  try(a <- DBI::dbExecute(connection, sql), T)
}

S7::method(add_index, db_mssql) <- function(connection, table, index, keys) {
  keys <- glue::glue_collapse(keys, sep = ", ")

  try(
    DBI::dbExecute(
      connection,
      glue::glue("CREATE INDEX {index} IF NOT EXISTS ON {table} ({keys});")
    ),
    T
  )
}

S7::method(add_index, db_postgres) <- function(connection, table, index, keys) {
  keys <- glue::glue_collapse(keys, sep = ", ")

  try(
    DBI::dbExecute(
      connection,
      glue::glue("CREATE INDEX IF NOT EXISTS {index} ON {table} ({keys});")
    ),
    T
  )
}

# drop_rows_where methods
S7::method(drop_rows_where, db_mssql) <- function(connection, table, condition) {
  t0 <- Sys.time()

  numrows <- DBI::dbGetQuery(connection, glue::glue(
    "SELECT COUNT(*) FROM {table} WHERE {condition};"
  )) |>
    as.numeric()

  num_deleting <- 100000
  num_deleting_character <- formatC(num_deleting, format = "f", drop0trailing = T)
  num_delete_calls <- ceiling(numrows / num_deleting)

  indexes <- csutil::easy_split(1:num_delete_calls, number_of_groups = 10)
  notify_indexes <- unlist(lapply(indexes, max))

  i <- 0
  while (numrows > 0) {
    b <- DBI::dbExecute(connection, glue::glue(
      "DELETE TOP ({num_deleting_character}) FROM {table} WHERE {condition}; ",
      "CHECKPOINT; "
    ))

    numrows <- DBI::dbGetQuery(connection, glue::glue(
      "SELECT COUNT(*) FROM {table} WHERE {condition};"
    )) |>
      as.numeric()
    i <- i + 1
  }

  t1 <- Sys.time()
  dif <- round(as.numeric(difftime(t1, t0, units = "secs")), 1)
}

S7::method(drop_rows_where, db_postgres) <- function(connection, table, condition) {
  t0 <- Sys.time()

  sql <- glue::glue("delete from {table} where {condition};")

  DBI::dbExecute(connection, sql)

  t1 <- Sys.time()
  dif <- round(as.numeric(difftime(t1, t0, units = "secs")), 1)
}

# keep_rows_where methods
S7::method(keep_rows_where, db_mssql) <- function(connection, table, condition, role_create_table = NULL) {
  t0 <- Sys.time()
  temp_name <- paste0("tmp", random_uuid())

  sql <- glue::glue("SELECT * INTO {temp_name} FROM {table} WHERE {condition}")
  DBI::dbExecute(connection, sql)

  DBI::dbRemoveTable(connection, name = table)

  sql <- glue::glue("EXEC sp_rename '{temp_name}', '{table}'")
  DBI::dbExecute(connection, sql)
  t1 <- Sys.time()
  dif <- round(as.numeric(difftime(t1, t0, units = "secs")), 1)
}

S7::method(keep_rows_where, db_postgres) <- function(connection, table, condition, role_create_table = NULL) {
  t0 <- Sys.time()
  temp_name <- paste0("tmp", random_uuid())

  sql <- glue::glue("SELECT * INTO {temp_name} FROM {table} WHERE {condition}")
  if(!is.na(role_create_table)) if(role_create_table!="x"){
    sql <- paste0("SET ROLE ", DBI::dbQuoteIdentifier(connection, role_create_table), "; ", sql,"; RESET ROLE")
  }
  DBI::dbExecute(connection, sql)

  sql <- glue::glue("DROP TABLE {table}")
  if(!is.na(role_create_table)) if(role_create_table!="x"){
    sql <- paste0("SET ROLE ", DBI::dbQuoteIdentifier(connection, role_create_table), "; ", sql,"; RESET ROLE")
  }
  DBI::dbExecute(connection, sql)

  sql <- glue::glue("ALTER TABLE {temp_name} RENAME TO {table}")
  if(!is.na(role_create_table)) if(role_create_table!="x"){
    sql <- paste0("SET ROLE ", DBI::dbQuoteIdentifier(connection, role_create_table), "; ", sql,"; RESET ROLE")
  }
  DBI::dbExecute(connection, sql)

  t1 <- Sys.time()
  dif <- round(as.numeric(difftime(t1, t0, units = "secs")), 1)
}

# drop_table methods
S7::method(drop_table, db_mssql) <- function(connection, table, role_create_table = NULL) {
  return(try(DBI::dbRemoveTable(connection, name = table), TRUE))
}

S7::method(drop_table, db_postgres) <- function(connection, table, role_create_table = NULL) {
  sql <- glue::glue("DROP TABLE {table}")
  if(!is.na(role_create_table)) if(role_create_table!="x"){
    sql <- paste0("SET ROLE ", DBI::dbQuoteIdentifier(connection, role_create_table), "; ", sql,"; RESET ROLE")
  }

  return(try(DBI::dbExecute(connection, sql), TRUE))
}

#' Covid-19 data for PCR-confirmed cases in Norway (nation and county)
#'
#' This data comes from the Norwegian Surveillance System for Communicable Diseases (MSIS).
#' The date corresponds to when the PCR-test was taken.
#'
#' The raw number of cases and cases per 100.000 population are recorded.
#'
#' This data was extracted on 2022-05-04.
#'
#' @format A csfmt_rts_data_v1 with 11028 rows and 18 variables:
#' \describe{
#'   \item{granularity_time}{day/isoweek}
#'   \item{granularity_geo}{nation, county}
#'   \item{country_iso3}{nor}
#'   \item{location_code}{norge, 11 counties}
#'   \item{border}{2020}
#'   \item{age}{total}
#'   \item{isoyear}{Isoyear of event}
#'   \item{isoweek}{Isoweek of event}
#'   \item{isoyearweek}{Isoyearweek of event}
#'   \item{season}{Season of event}
#'   \item{seasonweek}{Seasonweek of event}
#'   \item{calyear}{Calyear of event}
#'   \item{calmonth}{Calmonth of event}
#'   \item{calyearmonth}{Calyearmonth of event}
#'   \item{date}{Date of event}
#'   \item{covid19_cases_testdate_n}{Number of confirmed covid19 cases}
#'   \item{covid19_cases_testdate_pr100000}{Number of confirmed covid19 cases per 100.000 population}
#' }
#' @source \url{https://github.com/folkehelseinstituttet/surveillance_data/blob/master/covid19/_DOCUMENTATION_data_covid19_msis_by_time_location.txt}
"nor_covid19_cases_by_time_location"

# Auth hook ----

#' Set authentication hook for database connections
#'
#' @description
#' Register a function to be called when a database connection fails.
#' This is useful for refreshing Kerberos tickets or other authentication
#' credentials before retrying the connection.
#'
#' @param hook A function with no arguments that performs authentication,
#'   or NULL to clear the hook.
#' @return Invisibly returns the previous hook (if any).
#' @export
#' @examples
#' \dontrun{
#' # Set an auth hook to refresh Kerberos credentials
#' csdb_set_auth_hook(function() {
#'   system2("/bin/authenticate.sh", stdout = NULL)
#' })
#'
#' # Clear the hook
#' csdb_set_auth_hook(NULL)
#' }
csdb_set_auth_hook <- function(hook) {
  if (!is.null(hook) && !is.function(hook)) {
    stop("hook must be a function or NULL")
  }
  old_hook <- getOption("csdb.auth_hook")
  options(csdb.auth_hook = hook)
  invisible(old_hook)
}

#' Get the current authentication hook
#'
#' @description
#' Returns the currently registered authentication hook function.
#'
#' @return The current auth hook function, or NULL if none is set.
#' @export
csdb_get_auth_hook <- function() {
  getOption("csdb.auth_hook")
}

# DBConnection_v9 ----
#' R6 Class representing a database connection
#'
#' @description
#' A robust database connection manager that handles connections to various database systems
#' including Microsoft SQL Server and PostgreSQL. This class provides connection management,
#' authentication, and automatic reconnection capabilities.
#'
#' @details
#' The DBConnection_v9 class encapsulates database connection logic and provides a consistent
#' interface for connecting to different database systems. It supports both trusted connections
#' and user/password authentication, handles connection failures gracefully, and provides
#' automatic reconnection functionality.
#'
#' Key features:
#' \itemize{
#'   \item Support for multiple database systems (SQL Server, PostgreSQL)
#'   \item Automatic connection management with retry logic
#'   \item Secure credential handling
#'   \item Connection status monitoring
#'   \item Graceful error handling and recovery
#' }
#'
#' @import data.table
#' @import R6
#' @export DBConnection_v9
#' @examples
#' \dontrun{
#' # Create a SQL Server connection
#' db_config <- DBConnection_v9$new(
#'   driver = "ODBC Driver 17 for SQL Server",
#'   server = "localhost",
#'   port = 1433,
#'   db = "mydb",
#'   user = "myuser",
#'   password = "mypass"
#' )
#'
#' # Connect to the database
#' db_config$connect()
#'
#' # Check connection status
#' db_config$is_connected()
#'
#' # Use the connection
#' tables <- DBI::dbListTables(db_config$connection)
#'
#' # Disconnect when done
#' db_config$disconnect()
#'
#' # PostgreSQL example
#' pg_config <- DBConnection_v9$new(
#'   driver = "PostgreSQL",
#'   server = "localhost",
#'   port = 5432,
#'   db = "mydb",
#'   user = "myuser",
#'   password = "mypass"
#' )
#'
#' pg_config$connect()
#' # ... use connection ...
#' pg_config$disconnect()
#' }
DBConnection_v9 <- R6::R6Class(
  "DBConnection_v9",

  # public ----
  public = list(

    #' @field config Configuration details of the database.
    config = NULL,

    #' @description
    #' Create a new DBConnection_v9 object.
    #'
    #' @param driver Driver
    #' @param server Server
    #' @param port Port
    #' @param db DB
    #' @param schema Schema (e.g. "dbo")
    #' @param user User
    #' @param password Password
    #' @param trusted_connection NULL or "yes"
    #' @param sslmode NULL or "require"
    #' @param role_create_table NULL or the role to take when creating tables.
    #' @return A new `DBConnection_v9` object.
    initialize = function(
      driver = NULL,
      server = NULL,
      port = NULL,
      db = NULL,
      schema = NULL,
      user = NULL,
      password = NULL,
      trusted_connection = NULL,
      sslmode = NULL,
      role_create_table = NULL
      ) {
      force(driver)
      force(server)
      force(port)
      force(db)
      force(schema)
      force(user)
      force(password)
      force(trusted_connection)
      force(sslmode)
      force(role_create_table)

      if(is.null(trusted_connection)) trusted_connection <- "x"
      if(is.null(sslmode)) sslmode <- "x"
      if(is.null(role_create_table)) role_create_table <- "x"

      self$config <- list(
        driver = driver,
        server = server,
        port = port,
        db = db,
        schema = schema,
        user = user,
        password = password,
        trusted_connection = trusted_connection,
        sslmode = sslmode,
        role_create_table = role_create_table
      )
    },

    #' @description
    #' Is the DB schema connected?
    #' @return TRUE/FALSE
    is_connected = function() {
      retval <- FALSE
      if (is.null(private$pconnection)) {
        retval <- FALSE
      } else if (DBI::dbIsValid(private$pconnection)) {
        tryCatch({
          z <- private$pconnection |>
            DBI::dbListTables()
          retval <- TRUE
        }, error = function(e){
          retval <<- FALSE
        }, warning = function(e){
          retval <<- FALSE
        })
      }
      return(retval)
    },

    #' @description
    #' Class-specific print function.
    #' @param ... Not used.
    print = function(...) {
      if (!self$is_connected()) {
        if(requireNamespace("crayon", quietly = TRUE)) {
          cat(crayon::bgRed(crayon::white("(disconnected)\n\n")))
        } else {
          cat("(disconnected)\n\n")
        }
      } else {
        if(requireNamespace("crayon", quietly = TRUE)) {
          cat(crayon::bgCyan(crayon::white("(connected)\n\n")))
        } else {
          cat("(connected)\n\n")
        }
      }
      cat("Driver:             ", self$config$driver, "\n")
      cat("Server:             ", self$config$server, "\n")
      cat("Port:               ", self$config$port, "\n")
      cat("DB:                 ", self$config$db, "\n")
      cat("User:               ", self$config$user, "\n")
      cat("Password:           ", paste0(rep("*", nchar(self$config$password)), collapse=""), "\n")
      if(self$config$driver %in% c("PostgreSQL Unicode")) {
        cat("SSL mode:           ", self$config$sslmode, "\n")
      } else {
        cat("Trusted connection: ", self$config$trusted_connection, "\n")
      }
      cat("\n")

      invisible(self)
    },

    #' @description
    #' Connect to the database
    #' @param attempts Number of attempts to be made to try to connect
    connect = function(attempts = 2) {
      success <- FALSE
      auth_hook_called <- FALSE

      for(i in 1:attempts){
        tryCatch({
          private$connect_once()
          success <- TRUE
        },
        error = function(e){
          message("Attempt ", i,": ", e)
        })
        if(success) break()

        # If first attempt failed and we have an auth hook, call it
        if (i == 1 && !auth_hook_called) {
          auth_hook <- csdb_get_auth_hook()
          if (!is.null(auth_hook)) {
            message("Calling authentication hook...")
            tryCatch({
              auth_hook()
              auth_hook_called <- TRUE
            }, error = function(e) {
              message("Auth hook failed: ", conditionMessage(e))
            })
          }
        }

        # sleep to give the db time to recover
        # don't need to sleep on the last failed run
        if(i!=attempts) Sys.sleep(i)
      }
      if(!success) stop("Failed to connect to database after ", attempts, " attempts")
    },

    #' @description
    #' Disconnect from the database
    disconnect = function() {
      if(self$is_connected()) suppressWarnings(DBI::dbDisconnect(private$pconnection))
    }
  ),

  # active ----
  active = list(
    #' @field connection Database connection.
    connection = function(){
      private$pconnection
    },
    #' @field autoconnection Database connection that automatically connects if possible.
    autoconnection = function(){
      self$connect()
      return(private$pconnection)
    }
  ),

  # private ----
  private = list(
    pconnection = NULL,
    connect_once = function() {
      if(self$is_connected()){
        return()
      }

      # create connection
      tryCatch(
        {
          if (self$config$trusted_connection == "yes" & self$config$driver %in% c("ODBC Driver 17 for SQL Server")) {
            private$pconnection <- DBI::dbConnect(
              odbc::odbc(),
              driver = self$config$driver,
              server = self$config$server,
              port = self$config$port,
              trusted_connection = "yes"
            )
          } else if (self$config$driver %in% c("ODBC Driver 17 for SQL Server")) {
            private$pconnection <- DBI::dbConnect(
              odbc::odbc(),
              driver = self$config$driver,
              server = self$config$server,
              port = self$config$port,
              uid = self$config$user,
              pwd = self$config$password,
              encoding = "utf8"
            )
          } else if (self$config$sslmode == "require" & self$config$driver %in% c("PostgreSQL Unicode")) {
            private$pconnection <- DBI::dbConnect(
              odbc::odbc(),
              driver = self$config$driver,
              server = self$config$server,
              port = self$config$port,
              uid = self$config$user,
              password = self$config$password,
              database = self$config$db,
              sslmode = "require"
            )
          } else if (self$config$driver %in% c("PostgreSQL Unicode")) {
            private$pconnection <- DBI::dbConnect(
              odbc::odbc(),
              driver = self$config$driver,
              server = self$config$server,
              port = self$config$port,
              uid = self$config$user,
              password = self$config$password,
              database = self$config$db
            )
          } else {
            private$pconnection <- DBI::dbConnect(
              odbc::odbc(),
              driver = self$config$driver,
              server = self$config$server,
              port = self$config$port,
              user = self$config$user,
              password = self$config$password,
              encoding = "utf8"
            )
          }
        },
        error=function(cond){
          stop("Could not connect to database server '", self$config$server,"'\n",
               "Original error: ", conditionMessage(cond))
        }
      )

      # use db if available
      if(!is.null(self$config$db) & !self$config$driver %in% c("PostgreSQL Unicode")){
        tryCatch(
          {
            a <- DBI::dbExecute(private$pconnection, glue::glue({
              "USE {self$config$db};"
            }))
          },
          error = function(e) {
            stop("Database '", self$config$db,"' does not exist")
          }
        )
      }
    },
    finalize = function() {
      # message("Closing connection automatically")
      self$disconnect()
    }
  )
)

---
title: "Introduction to csdb"
author: "Richard Aubrey White"
date: "2025-07-18"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to csdb}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---




``` r
library(data.table)
library(magrittr)
```

## Concept

``` r
dbconnection <- csdb::DBConnection_v9$new(
  driver = Sys.getenv("CS9_DBCONFIG_DRIVER"),
  server = Sys.getenv("CS9_DBCONFIG_SERVER"),
  port = as.integer(Sys.getenv("CS9_DBCONFIG_PORT")),
  db = Sys.getenv("CS9_DBCONFIG_DB_ANON"),
  user = Sys.getenv("CS9_DBCONFIG_USER"),
  password = Sys.getenv("CS9_DBCONFIG_PASSWORD"),
  sslmode = Sys.getenv("CS9_DBCONFIG_SSLMODE")
)
dbconnection
#> (disconnected)
#> 
#> Driver:              PostgreSQL Unicode 
#> Server:              db 
#> Port:                5432 
#> DB:                  postgres 
#> User:                yourusername 
#> Password:            ********************* 
#> Trusted connection:  x
dbconnection$connect()

dbconnection$connection
#> <OdbcConnection> yourusername@db
#>   Database: postgres
#>   PostgreSQL Version: 17.0.5
dbconnection$autoconnection
#> <OdbcConnection> yourusername@db
#>   Database: postgres
#>   PostgreSQL Version: 17.0.5
dbconnection
#> (connected)
#> 
#> Driver:              PostgreSQL Unicode 
#> Server:              db 
#> Port:                5432 
#> DB:                  postgres 
#> User:                yourusername 
#> Password:            ********************* 
#> Trusted connection:  x

dbconnection$disconnect()
dbconnection$connection
#> Error: external pointer is not valid
class(dbconnection$connection)
#> [1] "PostgreSQL"
#> attr(,"package")
#> [1] "odbc"
class(dbconnection$autoconnection)
#> [1] "PostgreSQL"
#> attr(,"package")
#> [1] "odbc"

dbtable <- csdb::DBTable_v9$new(
  dbconfig = list(
    driver = Sys.getenv("CS9_DBCONFIG_DRIVER"),
    server = Sys.getenv("CS9_DBCONFIG_SERVER"),
    port = as.integer(Sys.getenv("CS9_DBCONFIG_PORT")),
    db = Sys.getenv("CS9_DBCONFIG_DB_ANON"),
    schema = Sys.getenv("CS9_DBCONFIG_SCHEMA_ANON"),
    user = Sys.getenv("CS9_DBCONFIG_USER"),
    password = Sys.getenv("CS9_DBCONFIG_PASSWORD"),
    sslmode = Sys.getenv("CS9_DBCONFIG_SSLMODE")
  ),
  table_name = "anon_test",
  field_types = c(
      "granularity_time" = "TEXT",
      "granularity_geo" = "TEXT",
      "country_iso3" = "TEXT",
      "location_code" = "TEXT",
      "border" = "INTEGER",
      "age" = "TEXT",
      "sex" = "TEXT",
      "isoyear" = "INTEGER",
      "isoweek" = "INTEGER",
      "isoyearweek" = "TEXT",
      "season" = "TEXT",
      "seasonweek" = "DOUBLE",
      "calyear" = "INTEGER",
      "calmonth" = "INTEGER",
      "calyearmonth" = "TEXT",
      "date" = "DATE",
      "covid19_cases_testdate_n" = "INTEGER",
      "covid19_cases_testdate_pr100000" = "DOUBLE"
    ),
    keys = c(
      "granularity_time",
      "location_code",
      "date",
      "age",
      "sex"
    ),
    indexes = list(
      "ind1" = c("granularity_time", "granularity_geo", "country_iso3", "location_code", "border", "age", "sex", "date", "isoyear", "isoweek", "isoyearweek")
    ),
    validator_field_types = csdb::validator_field_types_blank,
    validator_field_contents = csdb::validator_field_contents_blank
)
dbtable$drop_all_rows()
dbtable$insert_data(csdb::nor_covid19_cases_by_time_location)
dbtable$connect()
dbtable$dbconnection$is_connected()
#> [1] TRUE
dbtable$tbl()
#> # Source:   table<"public"."anon_test"> [?? x 18]
#> # Database: postgres  [yourusername@localhost:/postgres]
#>    granularity_time granularity_geo country_iso3 location_code border age   sex   isoyear isoweek isoyearweek season   seasonweek
#>    <chr>            <chr>           <chr>        <chr>          <int> <chr> <chr>   <int>   <int> <chr>       <chr>         <dbl>
#>  1 day              county          nor          county_nor03    2020 total total    2020       8 2020-08     2019/20…         31
#>  2 day              county          nor          county_nor03    2020 total total    2020       8 2020-08     2019/20…         31
#>  3 day              county          nor          county_nor03    2020 total total    2020       8 2020-08     2019/20…         31
#>  4 day              county          nor          county_nor03    2020 total total    2020       9 2020-09     2019/20…         32
#>  5 day              county          nor          county_nor03    2020 total total    2020       9 2020-09     2019/20…         32
#>  6 day              county          nor          county_nor03    2020 total total    2020       9 2020-09     2019/20…         32
#>  7 day              county          nor          county_nor03    2020 total total    2020       9 2020-09     2019/20…         32
#>  8 day              county          nor          county_nor03    2020 total total    2020       9 2020-09     2019/20…         32
#>  9 day              county          nor          county_nor03    2020 total total    2020       9 2020-09     2019/20…         32
#> 10 day              county          nor          county_nor03    2020 total total    2020       9 2020-09     2019/20…         32
#> # ℹ more rows
#> # ℹ 6 more variables: calyear <int>, calmonth <int>, calyearmonth <chr>, date <date>, covid19_cases_testdate_n <int>,
#> #   covid19_cases_testdate_pr100000 <dbl>
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{nor_covid19_cases_by_time_location}
\alias{nor_covid19_cases_by_time_location}
\title{Covid-19 data for PCR-confirmed cases in Norway (nation and county)}
\format{
A csfmt_rts_data_v1 with 11028 rows and 18 variables:
\describe{
  \item{granularity_time}{day/isoweek}
  \item{granularity_geo}{nation, county}
  \item{country_iso3}{nor}
  \item{location_code}{norge, 11 counties}
  \item{border}{2020}
  \item{age}{total}
  \item{isoyear}{Isoyear of event}
  \item{isoweek}{Isoweek of event}
  \item{isoyearweek}{Isoyearweek of event}
  \item{season}{Season of event}
  \item{seasonweek}{Seasonweek of event}
  \item{calyear}{Calyear of event}
  \item{calmonth}{Calmonth of event}
  \item{calyearmonth}{Calyearmonth of event}
  \item{date}{Date of event}
  \item{covid19_cases_testdate_n}{Number of confirmed covid19 cases}
  \item{covid19_cases_testdate_pr100000}{Number of confirmed covid19 cases per 100.000 population}
}
}
\source{
\url{https://github.com/folkehelseinstituttet/surveillance_data/blob/master/covid19/_DOCUMENTATION_data_covid19_msis_by_time_location.txt}
}
\usage{
nor_covid19_cases_by_time_location
}
\description{
This data comes from the Norwegian Surveillance System for Communicable Diseases (MSIS).
The date corresponds to when the PCR-test was taken.
}
\details{
The raw number of cases and cases per 100.000 population are recorded.

This data was extracted on 2022-05-04.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_table_names_and_info.R
\name{get_table_names_and_info}
\alias{get_table_names_and_info}
\title{Get table names, number of rows, and size information}
\usage{
get_table_names_and_info(connection)
}
\arguments{
\item{connection}{A database connection object (e.g., from \code{\link[DBI]{dbConnect}})}
}
\value{
A data.table containing table information with columns:
\describe{
  \item{table_name}{Character. Name of the table}
  \item{nrow}{Numeric. Number of rows in the table}
  \item{size_total_gb}{Numeric. Total size of the table in gigabytes}
  \item{size_data_gb}{Numeric. Size of data in gigabytes}
  \item{size_index_gb}{Numeric. Size of indexes in gigabytes}
}
}
\description{
Retrieves comprehensive information about database tables including their names,
row counts, and storage size metrics. This function provides database-specific
implementations for different database systems.
}
\examples{
\dontrun{
# Microsoft SQL Server example
con <- DBI::dbConnect(odbc::odbc(), 
                      driver = "ODBC Driver 17 for SQL Server",
                      server = "localhost", 
                      database = "mydb")
table_info <- get_table_names_and_info(con)
print(table_info)
DBI::dbDisconnect(con)

# PostgreSQL example  
con <- DBI::dbConnect(RPostgres::Postgres(),
                      host = "localhost",
                      dbname = "mydb",
                      user = "user")
table_info <- get_table_names_and_info(con)
print(table_info)
DBI::dbDisconnect(con)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/r6_DBConnection_v9.R
\name{csdb_get_auth_hook}
\alias{csdb_get_auth_hook}
\title{Get the current authentication hook}
\usage{
csdb_get_auth_hook()
}
\value{
The current auth hook function, or NULL if none is set.
}
\description{
Returns the currently registered authentication hook function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/r6_DBTable_v9.R
\name{validator_field_types_csfmt_rts_data_v1}
\alias{validator_field_types_csfmt_rts_data_v1}
\title{Field types validator for csfmt_rts_data_v1 schema}
\usage{
validator_field_types_csfmt_rts_data_v1(db_field_types)
}
\arguments{
\item{db_field_types}{A named character vector of database field types}
}
\value{
TRUE if field types are valid for csfmt_rts_data_v1, FALSE otherwise
}
\description{
Validates that field types conform to the csfmt_rts_data_v1 schema specification.
This validator ensures that the first 16 fields match the expected structure
for real-time surveillance data format version 1.
}
\examples{
# Valid field types for csfmt_rts_data_v1
valid_fields <- c(
  "granularity_time" = "TEXT",
  "granularity_geo" = "TEXT", 
  "country_iso3" = "TEXT",
  "location_code" = "TEXT",
  "border" = "INTEGER",
  "age" = "TEXT",
  "sex" = "TEXT",
  "isoyear" = "INTEGER",
  "isoweek" = "INTEGER",
  "isoyearweek" = "TEXT",
  "season" = "TEXT",
  "seasonweek" = "DOUBLE",
  "calyear" = "INTEGER",
  "calmonth" = "INTEGER",
  "calyearmonth" = "TEXT",
  "date" = "DATE",
  "cases_n" = "INTEGER"
)
validator_field_types_csfmt_rts_data_v1(valid_fields)

# Invalid field types (wrong structure)
invalid_fields <- c("id" = "INTEGER", "name" = "TEXT")
validator_field_types_csfmt_rts_data_v1(invalid_fields)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/r6_DBTable_v9.R
\name{validator_field_contents_csfmt_rts_data_v1}
\alias{validator_field_contents_csfmt_rts_data_v1}
\title{Field contents validator for csfmt_rts_data_v1 schema}
\usage{
validator_field_contents_csfmt_rts_data_v1(data)
}
\arguments{
\item{data}{A data.frame or data.table containing the data to validate}
}
\value{
TRUE if data is valid for csfmt_rts_data_v1, FALSE otherwise (with error attribute)
}
\description{
Validates that data contents conform to the csfmt_rts_data_v1 schema specification.
This validator checks that granularity_time and granularity_geo fields contain
valid values according to the surveillance data format requirements.
}
\examples{
# Valid data for csfmt_rts_data_v1
valid_data <- data.frame(
  granularity_time = c("date", "isoyearweek", "total"),
  granularity_geo = c("nation", "county", "municip"),
  stringsAsFactors = FALSE
)
validator_field_contents_csfmt_rts_data_v1(valid_data)

# Invalid data (wrong granularity_geo value)
invalid_data <- data.frame(
  granularity_time = "date",
  granularity_geo = "invalid_geo",
  stringsAsFactors = FALSE
)
validator_field_contents_csfmt_rts_data_v1(invalid_data)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/r6_DBConnection_v9.R
\name{csdb_set_auth_hook}
\alias{csdb_set_auth_hook}
\title{Set authentication hook for database connections}
\usage{
csdb_set_auth_hook(hook)
}
\arguments{
\item{hook}{A function with no arguments that performs authentication,
or NULL to clear the hook.}
}
\value{
Invisibly returns the previous hook (if any).
}
\description{
Register a function to be called when a database connection fails.
This is useful for refreshing Kerberos tickets or other authentication
credentials before retrying the connection.
}
\examples{
\dontrun{
# Set an auth hook to refresh Kerberos credentials
csdb_set_auth_hook(function() {
  system2("/bin/authenticate.sh", stdout = NULL)
})

# Clear the hook
csdb_set_auth_hook(NULL)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/r6_DBTable_v9.R
\name{validator_field_contents_blank}
\alias{validator_field_contents_blank}
\title{Blank data contents validator}
\usage{
validator_field_contents_blank(data)
}
\arguments{
\item{data}{A data.frame or data.table containing the data to validate}
}
\value{
Always returns TRUE
}
\description{
A pass-through validator that accepts any data without validation.
This is useful as a placeholder when no specific data content validation is needed.
}
\examples{
# This validator always returns TRUE regardless of input
test_data <- data.frame(id = 1:3, name = c("A", "B", "C"), value = c(10, 20, 30))
validator_field_contents_blank(test_data)

# Works with any data structure
empty_data <- data.frame()
validator_field_contents_blank(empty_data)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/r6_DBTable_v9.R
\name{DBTable_v9}
\alias{DBTable_v9}
\title{R6 Class representing a database table with advanced data management capabilities}
\description{
A comprehensive database table management class that provides high-level operations
for data manipulation, schema validation, and table administration. This class
combines database connectivity with data validation and efficient bulk operations.
}
\details{
The DBTable_v9 class is a sophisticated database table abstraction that provides:

\strong{Core functionality:}
\itemize{
  \item Table creation and schema management
  \item Data insertion with bulk loading capabilities
  \item Upsert operations (insert or update)
  \item Index management (creation, deletion)
  \item Data validation through customizable validators
  \item Integration with dplyr for data queries
}

\strong{Advanced features:}
\itemize{
  \item Automatic table creation based on field specifications
  \item Schema validation with custom validator functions
  \item Efficient bulk data loading using database-specific methods
  \item Index optimization for query performance
  \item Cross-database compatibility (SQL Server, PostgreSQL)
}

\strong{Data validation:}
The class supports custom validation functions for both field types and data contents,
ensuring data integrity and schema compliance.
}
\examples{
\dontrun{
# Create database connection
db_config <- list(
  driver = "ODBC Driver 17 for SQL Server",
  server = "localhost",
  db = "mydb",
  user = "myuser",
  password = "mypass"
)

# Define table schema
field_types <- c(
  "id" = "INTEGER",
  "name" = "TEXT",
  "value" = "DOUBLE",
  "date_created" = "DATE"
)

# Create table object
my_table <- DBTable_v9$new(
  dbconfig = db_config,
  table_name = "my_data_table",
  field_types = field_types,
  keys = c("id"),
  validator_field_types = validator_field_types_blank,
  validator_field_contents = validator_field_contents_blank
)

# Create table in database
my_table$create_table()

# Insert data
sample_data <- data.frame(
  id = 1:3,
  name = c("Alice", "Bob", "Charlie"),
  value = c(10.5, 20.3, 15.7),
  date_created = as.Date("2023-01-01")
)
my_table$insert_data(sample_data)

# Query data using dplyr
result <- my_table$tbl() |>
  dplyr::filter(value > 15) |>
  dplyr::collect()

# Add indexes for performance
my_table$add_indexes(c("name", "date_created"))

# Upsert (insert or update) data
new_data <- data.frame(
  id = 2:4,
  name = c("Bob_Updated", "Charlie", "David"),
  value = c(25.0, 15.7, 30.2),
  date_created = as.Date("2023-01-02")
)
my_table$upsert_data(new_data)
}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{dbconnection}}{Database connection.}

\item{\code{dbconfig}}{Configuration details of the database.}

\item{\code{table_name}}{Name of the table in the database.}

\item{\code{table_name_short_for_mssql_fully_specified_for_postgres}}{Fully specified name of the table in the database (e.g. \[db\].\[dbo\].\[table_name\]).}

\item{\code{table_name_short_for_mssql_fully_specified_for_postgres_text}}{Fully specified name of the table in the database (e.g. \[db\].\[dbo\].\[table_name\]).}

\item{\code{table_name_fully_specified}}{Fully specified name of the table in the database (e.g. \[db\].\[dbo\].\[table_name\]).}

\item{\code{table_name_fully_specified_text}}{Fully specified name of the table in the database (e.g. \[db\].\[dbo\].\[table_name\]) as a text string.}

\item{\code{field_types}}{The types of each column in the database table (INTEGER, DOUBLE, TEXT, BOOLEAN, DATE, DATETIME).}

\item{\code{field_types_with_length}}{The same as \code{field_types} but with \code{(100)} added to the end of all TEXT fields.}

\item{\code{keys}}{The combination of variables that uniquely identify each row in the database.}

\item{\code{keys_with_length}}{The same as \code{keys} but with \code{(100)} added to the end of all TEXT fields.}

\item{\code{indexes}}{A named list of vectors (generally "ind1", "ind2", etc.) that improves the speed of data retrieval operations on a database table.}

\item{\code{validator_field_contents}}{A function that validates the data before it is inserted into the database.}

\item{\code{load_folder}}{A temporary folder that is used to write data to before inserting into the database.}

\item{\code{censors}}{A named list of censors.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-DBTable_v9-new}{\code{DBTable_v9$new()}}
\item \href{#method-DBTable_v9-print}{\code{DBTable_v9$print()}}
\item \href{#method-DBTable_v9-connect}{\code{DBTable_v9$connect()}}
\item \href{#method-DBTable_v9-disconnect}{\code{DBTable_v9$disconnect()}}
\item \href{#method-DBTable_v9-table_exists}{\code{DBTable_v9$table_exists()}}
\item \href{#method-DBTable_v9-create_table}{\code{DBTable_v9$create_table()}}
\item \href{#method-DBTable_v9-remove_table}{\code{DBTable_v9$remove_table()}}
\item \href{#method-DBTable_v9-insert_data}{\code{DBTable_v9$insert_data()}}
\item \href{#method-DBTable_v9-upsert_data}{\code{DBTable_v9$upsert_data()}}
\item \href{#method-DBTable_v9-drop_all_rows}{\code{DBTable_v9$drop_all_rows()}}
\item \href{#method-DBTable_v9-drop_rows_where}{\code{DBTable_v9$drop_rows_where()}}
\item \href{#method-DBTable_v9-keep_rows_where}{\code{DBTable_v9$keep_rows_where()}}
\item \href{#method-DBTable_v9-drop_all_rows_and_then_upsert_data}{\code{DBTable_v9$drop_all_rows_and_then_upsert_data()}}
\item \href{#method-DBTable_v9-drop_all_rows_and_then_insert_data}{\code{DBTable_v9$drop_all_rows_and_then_insert_data()}}
\item \href{#method-DBTable_v9-tbl}{\code{DBTable_v9$tbl()}}
\item \href{#method-DBTable_v9-print_dplyr_select}{\code{DBTable_v9$print_dplyr_select()}}
\item \href{#method-DBTable_v9-add_indexes}{\code{DBTable_v9$add_indexes()}}
\item \href{#method-DBTable_v9-drop_indexes}{\code{DBTable_v9$drop_indexes()}}
\item \href{#method-DBTable_v9-confirm_indexes}{\code{DBTable_v9$confirm_indexes()}}
\item \href{#method-DBTable_v9-nrow}{\code{DBTable_v9$nrow()}}
\item \href{#method-DBTable_v9-info}{\code{DBTable_v9$info()}}
\item \href{#method-DBTable_v9-clone}{\code{DBTable_v9$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-new"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-new}{}}}
\subsection{Method \code{new()}}{
Create a new DBTable_v9 object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$new(
  dbconfig,
  table_name,
  field_types,
  keys,
  indexes = NULL,
  validator_field_types = validator_field_types_blank,
  validator_field_contents = validator_field_contents_blank
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{dbconfig}}{Configuration details of the database (driver, server, port, db, schema, user, password, trusted_connection, sslmode, role_create_table).}

\item{\code{table_name}}{Name of the table in the database.}

\item{\code{field_types}}{The types of each column in the database table (INTEGER, DOUBLE, TEXT, BOOLEAN, DATE, DATETIME).}

\item{\code{keys}}{The combination of these variables uniquely identifies each row of data in the table.}

\item{\code{indexes}}{A named list of vectors (generally "ind1", "ind2", etc.) that improves the speed of data retrieval operations on a database table.}

\item{\code{validator_field_types}}{A function that validates the \code{field_types} before the DB schema is created.}

\item{\code{validator_field_contents}}{A function that validates the data before it is inserted into the database.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new `DBTable_v9` object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-print"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-print}{}}}
\subsection{Method \code{print()}}{
Class-specific print function.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Not in use.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-connect"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-connect}{}}}
\subsection{Method \code{connect()}}{
Connect from the database
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$connect()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-disconnect"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-disconnect}{}}}
\subsection{Method \code{disconnect()}}{
Disconnect from the database
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$disconnect()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-table_exists"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-table_exists}{}}}
\subsection{Method \code{table_exists()}}{
Does the table exist
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$table_exists()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-create_table"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-create_table}{}}}
\subsection{Method \code{create_table()}}{
Create the database table
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$create_table()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-remove_table"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-remove_table}{}}}
\subsection{Method \code{remove_table()}}{
Drop the database table
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$remove_table()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-insert_data"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-insert_data}{}}}
\subsection{Method \code{insert_data()}}{
Inserts data
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$insert_data(
  newdata,
  confirm_insert_via_nrow = FALSE,
  verbose = TRUE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{newdata}}{The data to insert.}

\item{\code{confirm_insert_via_nrow}}{Checks nrow() before insert and after insert. If nrow() has not increased sufficiently, then attempt an upsert.}

\item{\code{verbose}}{Boolean.
Inserts data into the database table}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-upsert_data"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-upsert_data}{}}}
\subsection{Method \code{upsert_data()}}{
Upserts data into the database table
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$upsert_data(
  newdata,
  drop_indexes = names(self$indexes),
  verbose = TRUE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{newdata}}{The data to insert.}

\item{\code{drop_indexes}}{A vector containing the indexes to be dropped before upserting (can increase performance).}

\item{\code{verbose}}{Boolean.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-drop_all_rows"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-drop_all_rows}{}}}
\subsection{Method \code{drop_all_rows()}}{
Drops all rows in the database table
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$drop_all_rows()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-drop_rows_where"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-drop_rows_where}{}}}
\subsection{Method \code{drop_rows_where()}}{
Drops rows in the database table according to the SQL condition.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$drop_rows_where(condition)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{condition}}{SQL text condition.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-keep_rows_where"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-keep_rows_where}{}}}
\subsection{Method \code{keep_rows_where()}}{
Keeps rows in the database table according to the SQL condition.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$keep_rows_where(condition)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{condition}}{SQL text condition.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-drop_all_rows_and_then_upsert_data"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-drop_all_rows_and_then_upsert_data}{}}}
\subsection{Method \code{drop_all_rows_and_then_upsert_data()}}{
Drops all rows in the database table and then upserts data.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$drop_all_rows_and_then_upsert_data(
  newdata,
  drop_indexes = names(self$indexes),
  verbose = TRUE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{newdata}}{The data to insert.}

\item{\code{drop_indexes}}{A vector containing the indexes to be dropped before upserting (can increase performance).}

\item{\code{verbose}}{Boolean.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-drop_all_rows_and_then_insert_data"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-drop_all_rows_and_then_insert_data}{}}}
\subsection{Method \code{drop_all_rows_and_then_insert_data()}}{
Drops all rows in the database table and then inserts data.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$drop_all_rows_and_then_insert_data(
  newdata,
  confirm_insert_via_nrow = FALSE,
  verbose = TRUE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{newdata}}{The data to insert.}

\item{\code{confirm_insert_via_nrow}}{Checks nrow() before insert and after insert. If nrow() has not increased sufficiently, then attempt an upsert.}

\item{\code{verbose}}{Boolean.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-tbl"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-tbl}{}}}
\subsection{Method \code{tbl()}}{
Provides access to the database table via dplyr::tbl.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$tbl()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-print_dplyr_select"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-print_dplyr_select}{}}}
\subsection{Method \code{print_dplyr_select()}}{
Prints a template dplyr::select call that you can easily copy/paste for all your variables.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$print_dplyr_select()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-add_indexes"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-add_indexes}{}}}
\subsection{Method \code{add_indexes()}}{
Adds indexes to the database table from `self$indexes`
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$add_indexes()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-drop_indexes"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-drop_indexes}{}}}
\subsection{Method \code{drop_indexes()}}{
Drops all indees from the database table
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$drop_indexes()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-confirm_indexes"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-confirm_indexes}{}}}
\subsection{Method \code{confirm_indexes()}}{
Confirms that the names and number of indexes in the database are the same as in the R code.
Does not confirm the contents of the indexes!
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$confirm_indexes()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-nrow"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-nrow}{}}}
\subsection{Method \code{nrow()}}{
Gets the number of rows in the database table
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$nrow(use_count = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{use_count}}{If true, then uses the count command, which is slow but accurate. If false, then uses summary statistics, which is fast but inaccurate.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-info"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-info}{}}}
\subsection{Method \code{info()}}{
Gets the information about the database table
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$info()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBTable_v9-clone"></a>}}
\if{latex}{\out{\hypertarget{method-DBTable_v9-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBTable_v9$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/r6_DBTable_v9.R
\name{validator_field_types_blank}
\alias{validator_field_types_blank}
\title{Blank field types validator}
\usage{
validator_field_types_blank(db_field_types)
}
\arguments{
\item{db_field_types}{A named character vector of database field types}
}
\value{
Always returns TRUE
}
\description{
A pass-through validator that accepts any field types without validation.
This is useful as a placeholder when no specific field type validation is needed.
}
\examples{
# This validator always returns TRUE regardless of input
field_types <- c("id" = "INTEGER", "name" = "TEXT", "date" = "DATE")
validator_field_types_blank(field_types)

# Works with any field types
other_types <- c("value" = "DOUBLE", "status" = "BOOLEAN")
validator_field_types_blank(other_types)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/r6_DBConnection_v9.R
\name{DBConnection_v9}
\alias{DBConnection_v9}
\title{R6 Class representing a database connection}
\description{
A robust database connection manager that handles connections to various database systems
including Microsoft SQL Server and PostgreSQL. This class provides connection management,
authentication, and automatic reconnection capabilities.
}
\details{
The DBConnection_v9 class encapsulates database connection logic and provides a consistent
interface for connecting to different database systems. It supports both trusted connections
and user/password authentication, handles connection failures gracefully, and provides
automatic reconnection functionality.

Key features:
\itemize{
  \item Support for multiple database systems (SQL Server, PostgreSQL)
  \item Automatic connection management with retry logic
  \item Secure credential handling
  \item Connection status monitoring
  \item Graceful error handling and recovery
}
}
\examples{
\dontrun{
# Create a SQL Server connection
db_config <- DBConnection_v9$new(
  driver = "ODBC Driver 17 for SQL Server",
  server = "localhost",
  port = 1433,
  db = "mydb",
  user = "myuser",
  password = "mypass"
)

# Connect to the database
db_config$connect()

# Check connection status
db_config$is_connected()

# Use the connection
tables <- DBI::dbListTables(db_config$connection)

# Disconnect when done
db_config$disconnect()

# PostgreSQL example
pg_config <- DBConnection_v9$new(
  driver = "PostgreSQL",
  server = "localhost",
  port = 5432,
  db = "mydb",
  user = "myuser",
  password = "mypass"
)

pg_config$connect()
# ... use connection ...
pg_config$disconnect()
}
}
\section{Public fields}{
\if{html}{\out{<div class="r6-fields">}}
\describe{
\item{\code{config}}{Configuration details of the database.}
}
\if{html}{\out{</div>}}
}
\section{Active bindings}{
\if{html}{\out{<div class="r6-active-bindings">}}
\describe{
\item{\code{connection}}{Database connection.}

\item{\code{autoconnection}}{Database connection that automatically connects if possible.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-DBConnection_v9-new}{\code{DBConnection_v9$new()}}
\item \href{#method-DBConnection_v9-is_connected}{\code{DBConnection_v9$is_connected()}}
\item \href{#method-DBConnection_v9-print}{\code{DBConnection_v9$print()}}
\item \href{#method-DBConnection_v9-connect}{\code{DBConnection_v9$connect()}}
\item \href{#method-DBConnection_v9-disconnect}{\code{DBConnection_v9$disconnect()}}
\item \href{#method-DBConnection_v9-clone}{\code{DBConnection_v9$clone()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBConnection_v9-new"></a>}}
\if{latex}{\out{\hypertarget{method-DBConnection_v9-new}{}}}
\subsection{Method \code{new()}}{
Create a new DBConnection_v9 object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBConnection_v9$new(
  driver = NULL,
  server = NULL,
  port = NULL,
  db = NULL,
  schema = NULL,
  user = NULL,
  password = NULL,
  trusted_connection = NULL,
  sslmode = NULL,
  role_create_table = NULL
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{driver}}{Driver}

\item{\code{server}}{Server}

\item{\code{port}}{Port}

\item{\code{db}}{DB}

\item{\code{schema}}{Schema (e.g. "dbo")}

\item{\code{user}}{User}

\item{\code{password}}{Password}

\item{\code{trusted_connection}}{NULL or "yes"}

\item{\code{sslmode}}{NULL or "require"}

\item{\code{role_create_table}}{NULL or the role to take when creating tables.}
}
\if{html}{\out{</div>}}
}
\subsection{Returns}{
A new `DBConnection_v9` object.
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBConnection_v9-is_connected"></a>}}
\if{latex}{\out{\hypertarget{method-DBConnection_v9-is_connected}{}}}
\subsection{Method \code{is_connected()}}{
Is the DB schema connected?
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBConnection_v9$is_connected()}\if{html}{\out{</div>}}
}

\subsection{Returns}{
TRUE/FALSE
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBConnection_v9-print"></a>}}
\if{latex}{\out{\hypertarget{method-DBConnection_v9-print}{}}}
\subsection{Method \code{print()}}{
Class-specific print function.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBConnection_v9$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Not used.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBConnection_v9-connect"></a>}}
\if{latex}{\out{\hypertarget{method-DBConnection_v9-connect}{}}}
\subsection{Method \code{connect()}}{
Connect to the database
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBConnection_v9$connect(attempts = 2)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{attempts}}{Number of attempts to be made to try to connect}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBConnection_v9-disconnect"></a>}}
\if{latex}{\out{\hypertarget{method-DBConnection_v9-disconnect}{}}}
\subsection{Method \code{disconnect()}}{
Disconnect from the database
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBConnection_v9$disconnect()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-DBConnection_v9-clone"></a>}}
\if{latex}{\out{\hypertarget{method-DBConnection_v9-clone}{}}}
\subsection{Method \code{clone()}}{
The objects of this class are cloneable with this method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{DBConnection_v9$clone(deep = FALSE)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{deep}}{Whether to make a deep clone.}
}
\if{html}{\out{</div>}}
}
}
}
